#
# Copyright (c) 2017-2019,2024 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

# SOURCE_DIR = current source directory
# BINARY_DIR = current binary directory
# HG_ARCHIVAL = expected path to .hg_archival.txt
# INCLUDE_GUARD_NAME = name of #define to use as include guard

macro(gethgrev varname template)
	set(res_var "")
	execute_process(
		COMMAND ${HG_EXECUTABLE} log -r . --template ${template}
		WORKING_DIRECTORY ${SOURCE_DIR}
		OUTPUT_VARIABLE ${varname}
		RESULT_VARIABLE res_var
	)
	if(NOT ${res_var} EQUAL 0)
		unset(varname)
	endif()
endmacro()

if(EXISTS "${HG_ARCHIVAL}")
	set(tag "unknown")
	set(node "")
	set(dist 0)

	# read in the archival key-value pairs & pick out interesting ones
	file(STRINGS "${HG_ARCHIVAL}" blob)
	foreach(x ${blob})
		if("${x}" MATCHES "^node: ([0-9a-f]*)$")
			set(node ${CMAKE_MATCH_1})
		elseif("${x}" MATCHES "^latesttag: (.*)$")
			set(tag ${CMAKE_MATCH_1})
		elseif("${x}" MATCHES "^latesttagdistance: ([0-9]*)$")
			set(dist ${CMAKE_MATCH_1})
		endif()
	endforeach()

	# set HG_COMMIT_STR and VERSION to mimic the hg templates
	if(NOT ${node} STREQUAL "")
		set(HG_COMMIT_STR "${node}")
		if(${dist} EQUAL 0)
			set(VERSION "${tag}")
		else()
			string(SUBSTRING ${node} 0 12 shortnode)
			set(VERSION "${tag}+${dist}-${shortnode}")
		endif()
	endif()
else()
	find_package(Hg)

	if(HG_FOUND)
		gethgrev(HG_COMMIT_STR "{node}")

		# TODO: if there are uncommitted changes, append -dirty
		gethgrev(VERSION
			"{ifeq(latesttagdistance,0,\"{latesttag}\",\"{latesttag}+{latesttagdistance}-{node|short}\")}")
	else()
		message(FATAL_ERROR "Could not find Mercurial to get version info")
	endif()
endif()

if(NOT DEFINED HG_COMMIT_STR OR NOT DEFINED VERSION)
	message(FATAL_ERROR "Could not determine source code version information")
endif()

# convert ASCII hex commit hash to C byte array values
set(HG_COMMIT_BIN "")
set(HEX ${HG_COMMIT_STR})
while("${HEX}" MATCHES ".")
	string(SUBSTRING ${HEX} 0 2 COMMIT_BYTE)
	string(SUBSTRING ${HEX} 2 -1 HEX)
	set(HG_COMMIT_BIN "${HG_COMMIT_BIN}0x${COMMIT_BYTE},")
endwhile()

set(vstring "/* hgversion.h - Generated by cmake. Changes will be lost! */\n"
	"#ifndef ${INCLUDE_GUARD_NAME}\n"
	"#define ${INCLUDE_GUARD_NAME}\n"
	"#define HG_COMMIT_STR \"${HG_COMMIT_STR}\"\n"
	"#define HG_COMMIT_BIN { ${HG_COMMIT_BIN} }\n"
	"#define VERSION \"${VERSION}\"\n"
	"#endif\n")

file(WRITE hgversion.h.txt ${vstring})
# copy the file to the final header only if the version changes
# reduces needless rebuilds
execute_process(COMMAND ${CMAKE_COMMAND} -E copy_if_different
                        hgversion.h.txt ${BINARY_DIR}/hgversion.h)
execute_process(COMMAND ${CMAKE_COMMAND} -E remove hgversion.h.txt)
