/*
 * Copyright (c) 2020-2023 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

/*
 * ADIF 3.1.4
 *
 * http://www.adif.org/314/ADIF_314.htm
 */

#include <jeffpc/buffer.h>
#include <jeffpc/base64.h>
#include <jeffpc/nvl.h>
#include <jeffpc/error.h>

#include <hlog/adif.h>
#include <hlog/bands.h>
#include <hlog/time.h>
#include <hlog/util.h>
#include <hlog/version.h>

#include "qso-impl.h"

#define ADIF_VERSION "3.1.4"

#define HLOGPFX	"app_hlog_"

static struct val *__adif_qso_get_power(const struct qso_side *side);
static struct val *__adif_qso_get_freq(const struct qso_side *side);
static struct val *__adif_qso_get_band(const struct qso_side *side);
static struct val *__adif_qso_get_mode(const struct qso_side *side);
static struct val *__adif_qso_get_submode(const struct qso_side *side);
static struct val *__adif_qso_get_grid(const struct qso_side *side);
static struct val *__adif_qso_get_grid_ext(const struct qso_side *side);

struct side_field {
	const char *tx_name;
	const char *rx_name;
	struct val *(*get)(const struct qso_side *);
};

static const struct side_field side_fields[] = {
	{ "station_callsign", "call",             qso_get_station_call_val },
	{ "operator",         "contacted_op",     qso_get_operator_call_val },
	{ "tx_pwr",           "rx_pwr",           __adif_qso_get_power },
	{ "freq",             "freq_rx",          __adif_qso_get_freq },
	{ "band",             "band_rx",          __adif_qso_get_band },
	/*
	 * The mode
	 *
	 * Internally, we keep a single string mode per side of the QSO.
	 * ADIF 3.x splits a single QSO-level mode into two string - the
	 * mode and the submode.  Therefore, we possibly need to write out
	 * two strings to make the file compatible with other applications.
	 * However, because our mode has additional information that the
	 * ADIF mode/submode pair doesn't include (e.g., our DOMEX8 will map
	 * to DOMINO + DOMINOEX in ADIF), we want to write out our mode as
	 * well.
	 *
	 * This is the strict mode behavior.  Some consumers of ADIF
	 * logs seem to ignore the submode.  To allow interacting with them,
	 * we also allow storing the raw tx side mode into the mode ADIF
	 * field.  This is the relaxed behavior.
	 *
	 * Therefore, in strict mode, we generate:
	 *
	 *  - our mode as app-specific fields (hmode_[tr]x fields)
	 *  - tx mode as ADIF mode (mode field)
	 *  - tx mode as ADIF submode (submode field)
	 *
	 * And in relaxed mode, we generate:
	 *
	 *  - our mode as app-specific fields (hmode_[tr]x fields)
	 *  - our tx mode as ADIF mode (mode field)
	 *
	 * Note that we don't bother writing out the rx mode/submode, since
	 * nothing else would use it anyway.
	 *
	 * Note: The mode/submode generation is controlled by the
	 * strict_mode_fields/relaxed_mode_fields arrays below.
	 */
	{ HLOGPFX "hmode_tx", HLOGPFX "hmode_rx", qso_get_mode_val },
	{ "rst_sent",         "rst_rcvd",         qso_get_rst_val },
	{ "my_name",          "name",             qso_get_name_val },
	{ HLOGPFX "my_qth",    "qth",             qso_get_qth_val },
	{ "my_gridsquare",    "gridsquare",       __adif_qso_get_grid },
	{ "my_gridsquare_ext","gridsquare_ext",   __adif_qso_get_grid_ext },
	{ "my_country",       "country",          qso_get_country_val },
	{ "my_state",         "state",            qso_get_state_val },
	{ "my_cnty",          "cnty",             qso_get_county_val },
	{ HLOGPFX "my_cont",   "cont",            qso_get_continent_val },
	{ HLOGPFX "my_dxcc",   "dxcc",            qso_get_zone_dxcc },
	{ "my_itu_zone",      "ituz",             qso_get_zone_itu },
	{ "my_cq_zone",       "cqz",              qso_get_zone_cqz },
	{ "stx",              "srx",              qso_get_seq },
	{ "my_arrl_sect",     "arrl_sect",        qso_get_section_val },
};

static const struct side_field strict_mode_fields[] = {
	{ "mode",             NULL,               __adif_qso_get_mode },
	{ "submode",          NULL,               __adif_qso_get_submode },
};

static const struct side_field relaxed_mode_fields[] = {
	{ "mode",             NULL,               qso_get_mode_val },
};

static struct val *__adif_qso_get_power(const struct qso_side *side)
{
	char buf[32];

	if (!side->power)
		return NULL;

	if (side->power->i % 1000)
		snprintf(buf, sizeof(buf), "%"PRIu64".%03"PRIu64,
			 side->power->i / 1000,
			 side->power->i % 1000);
	else
		snprintf(buf, sizeof(buf), "%"PRIu64, side->power->i / 1000);

	return VAL_DUP_STR(buf);
}

static struct val *__adif_qso_get_freq(const struct qso_side *side)
{
	char buf[32];

	if (!pretty_print_freq(buf, sizeof(buf), side->freq))
		return NULL;

	return VAL_DUP_STR(buf);
}

static struct val *__adif_qso_get_band(const struct qso_side *side)
{
	if (!side->freq)
		return NULL;

	return bands_lookup_adif_name_val_by_freq(side->freq->i);
}

static struct val *__adif_qso_get_mode(const struct qso_side *side)
{
	struct str *val;

	if (!side->mode)
		return NULL;

	val = adif_mode_lookup_mode_by_hlog_mode(val_cstr(side->mode));

	return IS_ERR(val) ? NULL : str_cast_to_val(val);
}

static struct val *__adif_qso_get_submode(const struct qso_side *side)
{
	struct str *val;

	if (!side->mode)
		return NULL;

	val = adif_mode_lookup_submode_by_hlog_mode(val_cstr(side->mode));

	return IS_ERR(val) ? NULL : str_cast_to_val(val);
}

static struct val *__adif_qso_get_grid(const struct qso_side *side)
{
	struct str *full;
	struct str *tmp;

	full = qso_get_grid(side);
	if (!full)
		return NULL;

	if (str_len(full) <= 8)
		return str_cast_to_val(full);

	tmp = STR_DUP_LEN(str_cstr(full), 8);

	str_putref(full);

	return str_cast_to_val(tmp);
}

static struct val *__adif_qso_get_grid_ext(const struct qso_side *side)
{
	struct str *full;
	struct str *tmp;

	full = qso_get_grid(side);
	if (!full)
		return NULL;

	if (str_len(full) <= 8) {
		str_putref(full);
		return NULL;
	}

	tmp = STR_DUP(str_cstr(full) + 8);

	str_putref(full);

	return str_cast_to_val(tmp);
}

/*
 * Packing
 */

static int __adif_append(struct buffer *buf, const char *name,
			 const char *value, size_t len)
{
	char lenbuf[25];
	int ret;

	snprintf(lenbuf, sizeof(lenbuf), ":%zu>", len);

	if ((ret = buffer_append_c(buf, '<')))
		return ret;
	if ((ret = buffer_append_cstr(buf, name)))
		return ret;
	if ((ret = buffer_append_cstr(buf, lenbuf)))
		return ret;
	if (len && ((ret = buffer_append(buf, value, len))))
		return ret;
	if ((ret = buffer_append_c(buf, '\n')))
		return ret;

	return 0;
}

int adif_append_header(struct buffer *buf)
{
	int ret;

	if ((ret = buffer_append_cstr(buf, "###\n")))
		return ret;
	if ((ret = adif_append_kv_cstr(buf, "adif_ver", ADIF_VERSION)))
		return ret;
	if ((ret = adif_append_kv_cstr(buf, "programid", "hlog")))
		return ret;
	if ((ret = adif_append_kv_cstr(buf, "programversion", version_string)))
		return ret;
	if ((ret = buffer_append_cstr(buf, "<eoh>\n")))
		return ret;

	return 0;
}

int adif_append_kv_int(struct buffer *buf, const char *key, uint64_t value)
{
	char strval[32];

	snprintf(strval, sizeof(strval), "%"PRIu64, value);

	return __adif_append(buf, key, strval, strlen(strval));
}

int adif_append_kv_cstr(struct buffer *buf, const char *key, const char *value)
{
	return __adif_append(buf, key, value, strlen(value));
}

int adif_append_kv_time(struct buffer *buf, const char *key,
			const struct time *time)
{
	char strval[9];

	if (!(time->have & TIME_PART_TIME_HM))
		return 0;

	ASSERT3U(time->hour, <=, 23);
	ASSERT3U(time->min, <=, 59);

	if (time->have & TIME_PART_TIME_S)
		snprintf(strval, sizeof(strval), "%02u%02u%02u",
			 time->hour, time->min, time->sec);
	else
		snprintf(strval, sizeof(strval), "%02u%02u",
			 time->hour, time->min);

	return __adif_append(buf, key, strval, strlen(strval));
}

int adif_append_kv_date(struct buffer *buf, const char *key,
			const struct time *time)
{
	char strval[9];

	if (!(time->have & TIME_PART_DATE))
		return 0;

	ASSERT3U(time->year, <=, 9999);
	ASSERT3U(time->month, <=, 12);
	ASSERT3U(time->day, <=, 31);

	snprintf(strval, sizeof(strval), "%04u%02u%02u",
		 time->year, time->month, time->day);

	return __adif_append(buf, key, strval, strlen(strval));
}

int adif_append_kv_blob(struct buffer *buf, const char *key, const void *value,
			size_t length)
{
#define DATA_URI_PREFIX	"data:application/octet-stream;base64,"
	const size_t enclen = base64_required_length(length) +
		strlen(DATA_URI_PREFIX);
	char *encoded;

	encoded = alloca(enclen);

	strcpy(encoded, DATA_URI_PREFIX);
	base64_encode(encoded + strlen(DATA_URI_PREFIX), value, length);

	return __adif_append(buf, key, encoded, enclen);
}

int adif_append_kv(struct buffer *buf, const char *key, struct val *val)
{
	if (!val)
		return 0;

	switch (val->type) {
		case VT_INT:
			return adif_append_kv_int(buf, key, val->i);
		case VT_STR:
			return adif_append_kv_cstr(buf, key, val_cstr(val));
		case VT_BLOB:
			return adif_append_kv_blob(buf, key, val->blob.ptr,
						   val->blob.size);
		case VT_NULL:
		case VT_CHAR:
		case VT_BOOL:
		case VT_SYM:
		case VT_ARRAY:
		case VT_NVL:
		case VT_CONS:
			return -ENOTSUP;
	}

	panic("cannot set %s to unknown value type %d (%s)",
	      key, val->type, val_typename(val->type));
}

int adif_append_qso(struct buffer *buf, const struct qso *qso, bool with_header,
		    bool strict_mode)
{
	const struct side_field *mode_fields;
	const struct nvpair *pair;
	size_t num_mode_fields;
	size_t i;
	int ret;

	if (with_header) {
		ret = adif_append_header(buf);
		if (ret)
			return ret;
	}

#define CHECK(expr) \
	do { \
		ret = (expr); \
		if (ret) \
			return ret; \
	} while (0)

#define _KV(buf, k, v) \
	({ \
		struct val *val = (v); \
		int ret; \
		\
		ret = adif_append_kv((buf), (k), val); \
		val_putref(val); \
		ret; \
	})
#define KV(buf, k, v) CHECK(_KV(buf, k, v))
#define KV_CSTR(buf, k, v) CHECK(adif_append_kv_cstr((buf), (k), (v)))
#define KV_TIME(buf, k, v) CHECK(adif_append_kv_time((buf), (k), (v)))
#define KV_DATE(buf, k, v) CHECK(adif_append_kv_date((buf), (k), (v)))

	KV_CSTR(buf, "app_hlog_uuid", qso->uuid_str);
	KV_DATE(buf, "qso_date", &qso->start);
	KV_TIME(buf, "time_on", &qso->start);
	KV_DATE(buf, "qso_date_off", &qso->end);
	KV_TIME(buf, "time_off", &qso->end);
	KV(buf, "comment", qso_get_comment_val(qso));

	/* the ADIF mode/submode */
	if (strict_mode) {
		mode_fields = strict_mode_fields;
		num_mode_fields = ARRAY_LEN(strict_mode_fields);
	} else {
		mode_fields = relaxed_mode_fields;
		num_mode_fields = ARRAY_LEN(relaxed_mode_fields);
	}

	for (i = 0; i < num_mode_fields; i++) {
		const struct side_field *f = &mode_fields[i];

		/* always using the tx side */
		KV(buf, f->tx_name, f->get(&qso->tx));
	}

	/* per-side fields */
	for (i = 0; i < ARRAY_LEN(side_fields); i++) {
		const struct side_field *f = &side_fields[i];

		if (f->tx_name)
			KV(buf, f->tx_name, f->get(&qso->tx));
		if (f->rx_name)
			KV(buf, f->rx_name, f->get(&qso->rx));
	}

	nvl_for_each(pair, qso->additional)
		KV(buf, nvpair_name(pair), nvpair_value(pair));

#undef CHECK
#undef _KV
#undef KV
#undef KV_CSTR
#undef KV_TIME

	return buffer_append_cstr(buf, "<eor>\n");
}

struct buffer *qso_pack_adif(const struct qso *qso, bool with_header)
{
	struct buffer *buf;
	int ret;

	buf = buffer_alloc(1024);
	if (IS_ERR(buf))
		return buf;

	ret = adif_append_qso(buf, qso, with_header, true);

	if (ret)
		buffer_free(buf);

	return ret ? ERR_PTR(ret) : buf;
}
