/*
 * Copyright (c) 2020-2021 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/str.h>

#include <hlog/bands.h>

struct band {
	const char *name_iaru;
	const char *name_adif;
	uint64_t name_iaru_mm;
	uint64_t f_start;
	uint64_t f_end;

	/* val versions */
	struct val name_iaru_mm_val;
	struct str name_adif_val;
};

#define BAND(iaru, adif, mm, start, end) \
	{ \
		.name_iaru = (iaru), \
		.name_adif = (adif), \
		.name_iaru_mm = (mm), \
		.f_start = (start), \
		.f_end = (end), \
		.name_iaru_mm_val = INT_STATIC_INITIALIZER(mm), \
		.name_adif_val = STR_STATIC_INITIALIZER(adif), \
	}

static const struct band bands[] = {
	 BAND("23cm",  "23cm",      230, 1240000000, 1300000000),
	 BAND("33cm",  "33cm",      330,  902000000,  928000000),
	 BAND("70cm",  "70cm",      700,  420000000,  450000000),
	 BAND("1.25m", "1.25m",    1250,  219000000,  225000000),
	 BAND("2m",    "2m",       2000,  144000000,  148000000),
	 BAND("6m",    "6m",       6000,   50000000,   54000000),
	 BAND("10m",   "10m",     10000,   28000000,   29700000),
	 BAND("12m",   "12m",     12000,   24890000,   24990000),
	 BAND("15m",   "15m",     15000,   21000000,   21450000),
	 BAND("17m",   "17m",     17000,   18068000,   18168000),
	 BAND("20m",   "20m",     20000,   14000000,   14350000),
	 BAND("30m",   "30m",     30000,   10100000,   10150000),
	 BAND("40m",   "40m",     40000,    7000000,    7300000),
	 /* 60m band is channelized, but this is good enough */
	 BAND("60m",   "60m",     60000,    5330500,    5406500),
	 BAND("80m",   "80m",     80000,    3500000,    4000000),
	 BAND("160m",  "160m",   160000,    1800000,    2000000),
	 BAND("630m",  "630m",   630000,     472000,     479000),
	 BAND("2200m", "2190m", 2200000,     135700,     137800),
};

static const struct band *lookup_band_by_freq(uint64_t freq)
{
	size_t i;

	for (i = 0; i < ARRAY_LEN(bands); i++) {
		const struct band *band = &bands[i];

		if ((freq < band->f_start) || (freq > band->f_end))
			continue;

		return band;
	}

	return NULL;
}

const char *bands_lookup_name_by_freq(uint64_t freq)
{
	const struct band *band;

	band = lookup_band_by_freq(freq);

	return band ? band->name_iaru : NULL;
}

const char *bands_lookup_adif_name_by_freq(uint64_t freq)
{
	const struct band *band;

	band = lookup_band_by_freq(freq);

	return band ? band->name_adif : NULL;
}

struct val *bands_lookup_adif_name_val_by_freq(uint64_t freq)
{
	const struct band *band;

	band = lookup_band_by_freq(freq);
	if (!band)
		return NULL;

	/* cast away const - this is safe since static vals don't change */
	return str_cast_to_val((struct str *) &band->name_adif_val);
}

uint64_t bands_lookup_lambda_by_freq(uint64_t freq)
{
	const struct band *band;

	band = lookup_band_by_freq(freq);

	return band ? band->name_iaru_mm : 0;
}

struct val *bands_lookup_lambda_val_by_freq(uint64_t freq)
{
	const struct band *band;

	band = lookup_band_by_freq(freq);
	if (!band)
		return NULL;

	/* cast away const - this is safe since static vals don't change */
	return (struct val *) &band->name_iaru_mm_val;
}
