/*
 * Copyright (c) 2020-2023 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#ifndef __TIME_H
#define __TIME_H

#include <jeffpc/types.h>
#include <jeffpc/time.h>

enum time_parts {
	TIME_PART_DATE =	0x01, /* year, month, day */
	TIME_PART_TIME_HM =	0x02, /* hours, minutes */
	TIME_PART_TIME_S =	0x04, /* seconds */
#define TIME_PART_TIME_HMS	(TIME_PART_TIME_HM | TIME_PART_TIME_S)
};

struct time {
	uint16_t year;	/* 0-65535 */
	uint8_t month;	/* 1-12 */
	uint8_t day;	/* 1-31 */
	uint8_t hour;	/* 0-23 */
	uint8_t min;	/* 0-59 */
	uint8_t sec;	/* 0-60 */
	enum time_parts have:8;
};

extern int time_cmp(const struct time *a, const struct time *b);

/* Set to current date & time */
extern void time_set_now(struct time *time);

/*
 * Convert a string of the format "YYYYMMDD", "HHMMSS", or "HHMM" to a
 * struct time.
 */
extern int time_parse(struct time *time, const char *str, bool date);

/*
 * These print date/time and return true if the struct time contains
 * date/time, otherwise they fill in question marks and return false.
 */
extern bool time_snprint_date(char *buf, size_t bufsize, const struct time *t);
extern bool time_snprint_time(char *buf, size_t bufsize, const struct time *t);
/*
 * This prints the date/time in ISO8601/RFC3339 -style format and returns
 * true if the struct time contains date/time, otherwise it prints an empty
 * string and returns false.
 */
extern bool time_snprint_iso8601(char *buf, size_t bufsize, const struct time *t);

/*
 * Output a unix timestamp for the give time.  Note that unix timestamps do
 * not bother with leap seconds and therefore two different timestamps can
 * result in the same unix time.
 */
extern bool time_unix(const struct time *t, int64_t *out);

/* check that the date is valid */
extern bool valid_date(uint32_t y, uint32_t m, uint32_t d);
/* check that the time is valid */
extern bool valid_time(uint32_t h, uint32_t m, uint32_t s);

#endif
