/*
 * Copyright (c) 2015-2021,2024 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/select.h>
#include <ifaddrs.h>

#include <jeffpc/atomic.h>
#include <jeffpc/thread.h>
#include <jeffpc/error.h>
#include <jeffpc/types.h>
#include <jeffpc/time.h>
#include <jeffpc/io.h>
#include <jeffpc/mem.h>
#include <jeffpc/sock.h>

#include <hlog/udpsvc.h>
#include <hlog/config.h>

/*
 * This part of the library is a bit different.  It is meant to abstract
 * away the setup and teardown of a listening UDP server.  Regardless of
 * what the server is listening for, the same exact steps are needed to
 * start listening on a socket and accept connections.  The difference from
 * the rest of the common library is the fact that this code is not shy
 * about printing errors.
 */

#define MAX_SOCK_FDS		8
#define CONN_BACKLOG		64

struct udpsvc {
	pthread_t thread;
	void *private;
	void (*func)(void *, size_t, void *);

	struct udpsvc_rx {
		union xsockaddr addr;
		int fd;
	} rx[MAX_SOCK_FDS];
	int nrx;

	struct {
		union xsockaddr addr;
	} bcast[MAX_SOCK_FDS];
	int nbcast;

	struct udpsvc_tx {
		int fd;
		bool use;
		size_t addrlen;
	} inet, inet6;

	atomic_t server_shutdown;
};

static int bind_sock(struct udpsvc *state, int family, struct sockaddr *addr,
		     int addrlen)
{
	const int on = 1;
	int ret;
	int fd;

	if (state->nrx >= MAX_SOCK_FDS)
		return -EMFILE;

	fd = socket(family, SOCK_DGRAM, 0);
	if (fd == -1)
		return -errno;

	/* bind to IPv4 or IPv6 only, not both */
	setsockopt(fd, IPPROTO_IPV6, IPV6_V6ONLY, &on, sizeof(on));

	/* allow multiple programs to bind to the same <addr,port> */
	setsockopt(fd, SOL_SOCKET, SO_REUSEPORT, &on, sizeof(on));

	if (bind(fd, addr, addrlen))
		goto err;

	state->rx[state->nrx++].fd = fd;

	return 0;

err:
	ret = -errno;

	close(fd);

	return ret;
}

static int start_listening(struct udpsvc *state, const char *host,
			   uint16_t port)
{
	struct addrinfo hints, *res, *p;
	char strport[10];

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_DGRAM;
	hints.ai_flags = AI_PASSIVE;

	snprintf(strport, sizeof(strport), "%d", port);

	if (getaddrinfo(host, strport, &hints, &res))
		return -EINVAL; /* XXX: this should likely be something else */

	for (p = res; p; p = p->ai_next) {
		char str[XSOCKADDR_STRLEN];
		int ret;

		ASSERT3U(p->ai_family, ==, p->ai_addr->sa_family);

		switch (p->ai_family) {
			case AF_INET:
				state->inet.use = true;
				break;
			case AF_INET6:
				state->inet6.use = true;
				break;
			default:
				cmn_err(CE_INFO, "Unsupparted address family: %d",
					p->ai_family);
				continue;
		}

		xsockaddr_copy(&state->rx[state->nrx].addr, p->ai_addr);

		ret = bind_sock(state, p->ai_family, p->ai_addr, p->ai_addrlen);
		if (ret && ret != -EAFNOSUPPORT)
			return ret;
		else if (!ret)
			cmn_err(CE_INFO, "Bound to: %s %s (UDP)",
				xsockaddr_ntop(p->ai_addr, str, sizeof(str)),
				strport);
	}

	freeaddrinfo(res);

	return 0;
}

static void stop_listening(struct udpsvc *state)
{
	int i;

	for (i = 0; i < state->nrx; i++)
		close(state->rx[i].fd);
}

static int __start_broadcasting(struct udpsvc *state, struct udpsvc_tx *inet,
				int family)
{
	const int one = 1;
	int ret;
	int fd;

	if (!inet->use)
		return 0;

	fd = socket(family, SOCK_DGRAM, 0);
	if (fd < 0)
		return -errno;

	ret = setsockopt(fd, SOL_SOCKET, SO_BROADCAST, &one, sizeof(one));
	if (ret == -1) {
		ret = -errno;
		goto err;
	}

	inet->fd = fd;

	return 0;

err:
	close(fd);

	return ret;
}

static bool bcast_match(struct udpsvc *state, struct ifaddrs *ifa,
			const char *host)
{
	size_t i;

	for (i = 0; i < state->nbcast; i++)
		if (xsockaddr_cmp(ifa->ifa_broadaddr, &state->bcast[i].addr.sa))
			return false; /* already on the list */

	if (!host)
		return true; /* no host == all addresses & interfaces */

	for (i = 0; i < state->nrx; i++)
		if (xsockaddr_cmp(ifa->ifa_addr, &state->rx[i].addr.sa))
			return true;

	return false;
}

static int start_broadcasting(struct udpsvc *state, const char *host,
			      uint16_t port)
{
	struct ifaddrs *ifaddr;
	struct ifaddrs *cur;
	int ret;

	ret = __start_broadcasting(state, &state->inet, AF_INET);
	if (ret < 0)
		return ret;

	ret = __start_broadcasting(state, &state->inet6, AF_INET6);
	if (ret < 0)
		goto err;

	if (getifaddrs(&ifaddr) < 0)
		goto err;

	for (cur = ifaddr; cur; cur = cur->ifa_next) {
		union xsockaddr *dst = &state->bcast[state->nbcast].addr;
		struct sockaddr *sa = cur->ifa_addr;
		char str[XSOCKADDR_STRLEN];
		size_t len;
		void *addr;

		if (!sa || !cur->ifa_broadaddr || !cur->ifa_broadaddr->sa_len)
			continue;

		if (!bcast_match(state, cur, host))
			continue;

		if (state->nbcast >= MAX_SOCK_FDS) {
			ret = -EMFILE;
			goto err;
		}

		switch (sa->sa_family) {
			case AF_INET:
				if (!state->inet.use)
					continue;

				addr = &dst->inet.sin_addr;
				len = sizeof(dst->inet);
				break;
			case AF_INET6:
				if (!state->inet6.use)
					continue;

				addr = &dst->inet6.sin6_addr;
				len = sizeof(dst->inet6);
				break;
#ifdef HLOG_HAVE_AF_PACKET
			case AF_PACKET:
				continue;
#endif
			default:
				panic("Unknown address family %d",
				      sa->sa_family);
		}

		memcpy(dst, cur->ifa_broadaddr, len);

		switch (sa->sa_family) {
			case AF_INET:
				dst->inet.sin_port = htons(port);
				break;
			case AF_INET6:
				dst->inet6.sin6_port = htons(port);
				break;
		}

		state->nbcast++;

		inet_ntop(dst->sa.sa_family, addr, str, sizeof(str));

		cmn_err(CE_INFO, "Broadcasting to %s %u (UDP) on %s",
			str, port, cur->ifa_name);
	}

	freeifaddrs(ifaddr);

	return 0;

err:
	if (state->inet.fd >= 0)
		close(state->inet.fd);
	if (state->inet6.fd >= 0)
		close(state->inet6.fd);

	return ret;
}

static void stop_broadcasting(struct udpsvc *state)
{
	if (state->inet.fd >= 0)
		close(state->inet.fd);
	if (state->inet6.fd >= 0)
		close(state->inet6.fd);
}

static void *accept_conns(void *_state)
{
	struct udpsvc *state = _state;
	fd_set set;
	int maxfd;
	int ret;
	int i;

	for (;;) {
		union xsockaddr addr;
		unsigned len;

		FD_ZERO(&set);
		maxfd = 0;
		for (i = 0; i < state->nrx; i++) {
			FD_SET(state->rx[i].fd, &set);
			maxfd = MAX(maxfd, state->rx[i].fd);
		}

		ret = select(maxfd + 1, &set, NULL, NULL, NULL);

		if (atomic_read(&state->server_shutdown))
			break;

		if ((ret < 0) && (errno != EINTR))
			cmn_err(CE_ERROR, "Error on select: %s",
				strerror(errno));

		for (i = 0; (i < state->nrx) && (ret > 0); i++) {
			uint8_t buf[10240];
			int bytes;

			if (!FD_ISSET(state->rx[i].fd, &set))
				continue;

			len = sizeof(addr);
			bytes = recvfrom(state->rx[i].fd, buf, sizeof(buf), 0,
					 &addr.sa, &len);
			if (bytes < 0) {
				cmn_err(CE_INFO, "udpsvc: failed to recv: %s",
					strerror(errno));
				continue;
			}

			state->func(buf, bytes, state->private);

			/* XXX: should this be executed every time we loop? */
			ret--;
		}
	}

	return NULL;
}

struct udpsvc *udpsvc(const char *host, uint16_t port, bool enable_bcast,
		      const char *threadname,
		      void (*func)(void *, size_t, void *), void *private)
{
	struct udpsvc *state;
	int ret;

	state = zalloc(sizeof(struct udpsvc));
	if (!state)
		return ERR_PTR(-ENOMEM);

	state->private = private;
	state->func = func;
	state->inet.fd = -1;
	state->inet.addrlen = sizeof(struct sockaddr_in);
	state->inet6.fd = -1;
	state->inet6.addrlen = sizeof(struct sockaddr_in6);
	atomic_set(&state->server_shutdown, false);

	ret = start_listening(state, host, port);
	if (ret)
		goto err;

	if (enable_bcast) {
		ret = start_broadcasting(state, host, port);
		if (ret)
			goto err_close;
	}

	ret = xthr_create_named(&state->thread, threadname, accept_conns, state);
	if (ret)
		goto err_close2;

	return state;

err_close2:
	stop_broadcasting(state);

err_close:
	stop_listening(state);

err:
	free(state);

	return ERR_PTR(ret);
}

void udpsvc_stop(struct udpsvc *state)
{
	int ret;

	if (!state)
		return;

	atomic_set(&state->server_shutdown, true);

	ret = xthr_join(state->thread, NULL);
	if (ret)
		cmn_err(CE_ERROR, "%s: failed to join thread: %s\n", __func__,
			xstrerror(ret));

	stop_listening(state);

	free(state);
}

int udpsvc_bcast(struct udpsvc *state, const void *data, size_t len)
{
	ssize_t ret;
	size_t i;

	if (!state->nbcast)
		return -ENOTSUP;

	for (i = 0; i < state->nbcast; i++) {
		struct udpsvc_tx *tx;

		switch (state->bcast[i].addr.sa.sa_family) {
			case AF_INET:
				tx = &state->inet;
				break;
			case AF_INET6:
				tx = &state->inet6;
				break;
			default:
				panic("Unknown address family %d",
				      state->bcast[i].addr.sa.sa_family);
		}

		ret = sendto(tx->fd, data, len, 0, &state->bcast[i].addr.sa,
			     tx->addrlen);
		if (ret != len) {
			char str[XSOCKADDR_STRLEN];

			cmn_err(CE_ERROR, "Failed to broadcast to %s (idx %zu): "
				"%zd of %zu bytes sent",
				xsockaddr_ntop(&state->bcast[i].addr.sa, str,
					       sizeof(str)),
				i, ret, len);
		}
	}

	return 0;
}
