/*
 * Copyright (c) 2021-2024 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

/*
 * The official definition of CQ zones used to be at:
 * https://cq-amateur-radio.com/cq_awards/cq_waz_awards/cq_waz_list.html
 *
 * The content of the page as of 2023-05-02 was:
 *
 * -------------------------------------8<-------------------------------------
 * Section 8. WAZ Zone/Country/Entities List
 *
 * Zone 1. Northwestern Zone of North America: KL (Alaska), VY1 Yukon, VE8 the
 * Northwest and VY0 Nunavut Territories west of 102 degrees (Includes the
 * islands of Victoria, Banks, Melville, and Prince Patrick).
 *
 * Zone 2. Northeastern Zone of North America: VO2 (Labrador), the portion of
 * VE2 Quebec north of the 50th parallel, and Nunavut Territories east of 102
 * degrees (Includes the islands of King Christian, King William. Prince of
 * Wales, Somerset, Bathurst, Devon, Ellesmere, Baffin and the Melville and
 * Boothia Peninsulas, excluding Akimiski Island, Bear Islands and East Pen
 * Island in Hudson Bay).
 *
 * Zone 3. Western Zone of North America: VE7 (British Columbia), W6, and the
 * W7 states of Arizona, Idaho, Nevada, Oregon, Utah, and Washington.
 *
 * Zone 4. Central Zone of North America: VE3 (Ontario), VE4 (Manitoba), VE5
 * (Saskatchewan), VE6 (Alberta), VY0 Akimiski Island, and Bear Islands, and
 * Fox Island and East Pen Island in Hudson Bay. The W7 states of Montana and
 * Wyoming, W0, W9, W8 (except West Virginia), W5, and the W4 states of
 * Alabama, Tennessee, and Kentucky.
 *
 * Zone 5. Eastern Zone of North America: 4U1UN, CY9 (St. Paul Is.), CY0 (Sable
 * Is.), FP (St. Pierre Miquelon), VE1 (Nova Scotia) and VE9 (New Brunswick),
 * VY2 (Prince Edward Is.), VO1 (Newfoundland) and the portion of VE2 Quebec
 * south of the 50th parallel. VP9 (Bermuda), W1, W2, W3 and the W4 states of
 * Florida, Georgia, South Carolina, North Carolina, Virginia and the W8 state
 * of West Virginia.
 *
 * Zone 6. Southern Zone of North America: XE (Mexico), XF4 (Revilla Gigedo).
 *
 * Zone 7. Central American Zone: FO (Clipperton), HK0 (San Andres Is.), HP
 * (Panama), HR (Honduras), TG (Guatemala), TI (Costa Rica), TI9 (Cocos Is.),
 * V3 (Belize), YN (Nicaragua) and YS (El Salvador).
 *
 * Zone 8. West Indies Zone: C6 (Bahamas), CO (Cuba), FG (Guadeloupe),  FJ (St.
 * Barts), FM (Martinique), FS (Saint Martin), HH (Haiti), HI (Dominican
 * Republic), J3 (Grenada), J6 (St. Lucia), J7 (Dominica), J8 (St. Vincent),
 * KG4 (Guantanamo Base), KP1 (Navassa Is.), KP2 Virgin Islands), KP4 (Puerto
 * Rico), KP5 (Navassa Is.), PJ5 (Eustatius), PJ6 (Saba), PJ7 (Sint Maarten).
 * V2 (Antigua and Barbuda), V4 (St. Kitts and Nevis), VP2E (Anguilla), VP2M
 * (Montserrat), VP2V (British Virgin Is.), VP5 (Turks and Caicos Is.), YV0
 * (Aves Is.), ZF (Cayman Is.), 6Y (Jamaica), and 8P (Barbados).
 *
 * Zone 9. Northern Zone of South America: FY (French Guyana), HK (Colombia),
 * HK0 (Malpelo Is.), PJ2 (Curacao), PJ4 (Bonaire), PZ (Surinam), YV
 * (Venezuela), 8R (Guiana),P4 Aruba and 9Y (Trinidad and Tobago Is.).
 *
 * Zone 10. Western Zone of South America: CP (Bolivia), HC (Ecuador), HC8
 * (Galapagos Is.), and OA (Peru).
 *
 * Zone 11. Central Zone of South America: PY (Brazil), PY0 (Fernando de
 * Noronha, PY0 (St. Peter and Paul Rocks) and PY0 Trinidad Martin Vaz, ZP
 * (Paraguay).
 *
 * Zone 12. Southwest Zone of South America: CE (Chile), CE0 (Easter Is.),
 * CE0(Juan Fernandez Is.), CE0 (San Felix Is.), 3Y (Peter I) and some
 * Antarctic stations (See Notes Below).
 *
 * Zone 13. Southeast Zone of South America: CX (Uruguay), LU (Argentina), VP8
 * Falklands, VP8 South Georgia, VP8 South Orkney, VP8 South Shetlands, VP8
 * South Sandwich  and some Antarctic stations (See Notes Below)
 *
 * Zone 14. Western Zone of Europe: C3 (Andorra), CT (Portugal), CU (Azores
 * Is.), DL (Germany), EA (Spain), EA6 (Balearic Is.), El (Ireland), F
 * (France), G (England), GD (Isle of Man), GI (Northern Ireland), GJ (Jersey),
 * GM (Scotland), GU (Guernsey), GW (Wales), HB (Switzerland),
 * HB0(Liechtenstein), LA (Norway), LX (Luxembourg), ON (Belgium), OY (Faroe
 * Is.), OZ (Denmark), PA (Netherlands), SM (Sweden), ZB (Gibraltar), 3A
 * (Monaco) and 4U1ITU (ITU, Geneva).
 *
 * Zone 15. Central European Zone: E7 (Bosnia Herzegovina), ES (Estonia), HA
 * (Hungary), HV (Vatican), I (Italy), IS0 (Sardinia), LY (Lithuania), OE
 * (Austria), OH (Finland), OH0 (Aland Is.), OJ0 (Market Reef), OK (Czech
 * Rep.), OM (Slovakia), S5 (Slovenia), SP (Poland), T7 (San Marino),  TK
 * (Corsica), UA2,F,K, RA2,UB2-UI2 (Kaliningrad), YL (Latvia), YU (Serbia), ZA
 * (Albania), Z3 (Macedonia), 1A0 (SMOM), 4O (Montenegro), 9A (Croatia), 9H
 * (Malta).
 *
 * Zone 16. Eastern Zone of Europe: EW (Belarus), ER (Moldova), UA1, UA2(except
 * for RA2 and UA2-UI2), UA3. UA4,  UA5,UA6, UA7,UA9,UA9 (S,T,W), and UR
 * (Ukraine).
 *
 * Zone 17. Western Zone of Siberia: EX (Kirgizstan), EY (Tajikistan), EZ
 * (Turkmenistan), UA8,9 (A, B,C,D, F, G, J. K, L, M, N, Q,R, X), UK
 * (Uzbekistan), UN (Kazakhstan).
 *
 * Zone 18. Central Siberian Zone: UA8,9 (H,I,O,P,U, V,Y,Z), and UAO (A, B, H,
 * O,R, S,T, U,V, W)
 *
 * Zone 19. Eastern Siberian Zone: UA0 (C, D, F, I, J, K, L, Q, X, Z)
 *
 * Zone 20. Balkan Zone: E4 (Palestine), JY (Jordan), LZ (Bulgaria), OD
 * (Lebanon), SV (Greece), SV/A (Mt. Athos), SV5 (Dodecanese), SV9 (Crete), TA
 * (Turkey), YK (Syria), YO (Romania), ZC4 (UK Sov. Base), 4X (Israel) and 5B
 * (Cyprus).
 *
 * Zone 21. Southwestern Zone of Asia: A4 (Oman), A6 (United Arab Emirates), A7
 * (Qatar), A9 (Bahrain), AP (Pakistan), EK (Armenia), EP (Iran), HZ (Saudi
 * Arabia), YA (Afghanistan), YI (Iraq), 4J (Azerbaijan), 4L (Georgia), 7O
 * (excluding Socotra I. and Abd al Kuri Is.) and 9K (Kuwait).
 *
 * Zone 22. Southern Zone of Asia: A5 (Bhutan), S2 (Bangladesh), VU (India), VU
 * (Lakshadweep Is.), 4S (Sri Lanka), 8Q (Maldives.), and 9N (Nepal).
 *
 * Zone 23. Central Zone of Asia: JT (Mongolia), UA0Y, BY3G‑L (NeiMenggu),
 * BY9G‑L (GanSu), BY9M‑R (NingXia), BY9S‑Z (QingHai) and BY0.
 *
 * Zone 24. Eastern Zone of Asia: BV9P (Pratas Is.), BV (Taiwan), BY1, BY2,
 * BY3A‑F (TianJin), BY3M‑R (HeBei), BY3S-X (ShanXi), BY4, BY5, BY6, BY7, BY8,
 * BY9A‑F (ShaanXi), VR (Hong Kong), and XX (Macau).
 *
 * Zone 25. Japanese Zone: HL (South Korea), JA (Japan), and P5 (North Korea).
 *
 * Zone 26. Southeastern Zone of Asia:  * (Paracel Islands), 3W (Vietnam), E2
 * (Thailand), VU (Andaman and Nicobar Islands), XU (Cambodia), XW (Laos), XZ
 * (Myanmar), and 1S/9M0 (Spratly Islands)
 *
 * Zone 27. Philippine Zone: BS7 (Scarborough Reef), DU (Philippines), JD1
 * (Minami Torishima), JD1 (Ogasawara), T8 (Palau), KH2 (Guam), KH0 (Mariana
 * Is.), V6 (Fed. States of Micronesia).
 *
 * Zone 28. Indonesian Zone: H4 (Solomon Is.), P2 (Papua New Guinea), V8
 * (Brunei), YB (Indonesia), 4W (East Timor), 9M (West and East Malaysia) and
 * 9V (Singapore).
 *
 * Zone 29. Western Zone of Australia: VK6 (Western Australia), VK8 (Northern
 * Territory), VK9X (Christmas Is.), VK9C (Cocos‑Keeling Is.) and some
 * Antarctic stations :(See Notes Below)
 *
 * Zone 30. Eastern Zone of Australia: FK/C (Chesterfield Is.), VK1 (Capital
 * Territory), VK2 (New South Wales), VK3 (Victoria) and VK4 (Queensland), VK5
 * (South Australia), VK7 (Tasmania), VK9L (Lord Howe Is.), VK9M (Mellish
 * Reef), VK9 (Willis Is.), VK0M (Macquarie Is.) and some Antarctic stations
 * (See Notes Below).
 *
 * Zone 31. Central Pacific Zone: C2 (Nauru), FO (Marquesas), KH1 (Baker
 * Howland Is.), KH3 (Johnson Is.), KH4 (Midway Is.), KH5K (Kingman Reef), KH5
 * (Palmyra Jarvis), KH6 (Hawaii), KH7K(Kure), KH9 (Wake Is), T2 (Tuvalu), T30
 * (Western Kiribati), T31 (Central Kiribati), and T32 (Eastern Kiribati),T33
 * (Bananba Is), V7 (Marshall Is.), and ZK3 (Tokelau)
 *
 * Zone 32. New Zealand Zone: A3 (Tonga), E5 (North and South Cook Is.) FK New
 * Caledonia but NOT Chesterfield Is.), FO (French Polynesia and Austral Is.
 * but NOT Marquesas and Clipperton), FW (Wallis Is.), H40(Temotu), KH8
 * (American Samoa), KH8S (Swains), VK9N (Norfolk Is.), VP6 (Pitcairn and
 * Ducie), YJ (Vanuatu), ZK2 (Niue Is.), ZL (New Zealand), ZL7 (Chatham Is.),
 * ZL8 (Kermadec Is.), ZL9 (Auckland and Campbell Is.), 3D2 (Fiji, Rotuma, and
 * Conway Reef), 5W (Western Samoa) and some Antarctic stations (See Notes
 * Below).
 *
 * Zone 33. Northwestern Zone of Africa: CN (Morocco), CT3 (Madeira Is), EA8
 * (Canary Is.), EA9 (Ceuta and/or Melilla), IG9 (Pelagic Is.), IH9 IG9
 * (Pantelleria Is. or Pelagic Islands), S0 (Western Sahara), 3V (Tunisia), and
 * 7X (Algeria).
 *
 * Zone 34. Northeastern Zone of Africa: ST (Sudan), SU (Egypt), Z8 (South
 * Sudan) and 5A (Libya).
 *
 * Zone 35. Central Zone of Africa: C5 (The Gambia), D4 (Cape Verde Is.), EL
 * (Liberia), J5 (Guinea Bissau), TU (Cote d'Ivoire), TY (Benin), TZ (Mali), XT
 * (Burkina Faso), 3X (Guinea), 5N (Nigeria), 5T (Mauritania), 5U (Niger), 5V
 * (Togo), 6W (Senegal), 9G (Ghana) and 9L (Sierra Leone).
 *
 * Zone 36. Equatorial Zone of Africa: D2 (Angola), S9 (Sao Tome Principe), TJ
 * (Cameroon), TL (Central African Rep.), TN (Congo), TR (Gabon), TT (Chad),
 * ZD7 (St. Helena Is.), ZD8 (Ascension Is.), 3C (Equatorial Guinea), 3C0
 * (Annobon Is.), 9J (Zambia), 9Q (Democratic Rep. of Congo), 9U (Burundi) and
 * 9X (Rwanda).
 *
 * Zone 37. Eastern Zone of Africa: C9 (Mozambique), ET (Ethiopia), E3
 * (Eritrea), J2 (Djibouti), T5 (Somalia), 5H (Tanzania), 5X (Uganda), 5Z
 * (Kenya), 7O (Socotra and Abd al Kuri islands ONLY) and 7Q (Malawi).
 *
 * Zone 38. South African Zone: A2 (Botswana), V5 (Namibia), ZD9 (Tristan da
 * Cunha Gough Is.), Z2 (Zimbabwe), ZS (South Africa) and ZS8 (Marion Is), 3DA
 * (Swaziland), 3Y (Bouvet Is.), 7P (Lesotho), and some Antarctic stations (See
 * Notes Below)
 *
 * Zone 39. Madagascar Zone: D6 (Comoros), FH (Mayotte), FR (Reunion),FT-G
 * (Glorioso), FT-J&E (Juan de Nova and Europa), FT-T (Tromelin),  FT‑W (Crozet
 * Is.), FT-X (Kerguelen Is.), FT‑Z (Amsterdam St. Paul Is.), S7 (Seychelles),
 * VK0 (Heard Is.), VQ9 (Chagos), 3B6/7 (Agalega St.Brandon Is.), 3B8
 * (Mautitius Is.), 3B9 (Rodriguez Is.), 5R (Madagascar) and some Antarctic
 * stations (See Notes Below)
 *
 * Zone 40. North Atlantic Zone: JW (Svalbard), JX (Jan Mayen), OX (Greenland),
 * RI1FJ (Franz Josef Land), and TF (Iceland).
 *
 * (*) The Paracel Islands have no official or unofficial prefix
 *
 * Antarctic notes:
 * The boundaries of CQ zones 12, 13, 29, 30, 32, 38 and 39 converge at the
 * South Pole.
 *
 * Stations KC4AAA and KC4USN are at the South Pole, and will count for any one
 * of the listed zones. Most Antarctic stations indicate their zone on the QSL
 * card.
 *
 * Updated and correct as of April 1, 2018
 * -------------------------------------8<-------------------------------------
 */

#include "zones_impl.h"

static const struct zone canada[13] = {
	{ "NS", ZP_CONST, .constant = 5 },
	{ "QC", ZP_NORTHOF,
		.latlon.boundary = 50.,
		.latlon.zones = { 5, 2 } },
	{ "ON", ZP_CONST, .constant = 4 },
	{ "MB", ZP_CONST, .constant = 4 },
	{ "SK", ZP_CONST, .constant = 4 },
	{ "AB", ZP_CONST, .constant = 4 },
	{ "BC", ZP_CONST, .constant = 3 },
	{ "NT", ZP_CONST, .constant = 1 },
	{ "NB", ZP_CONST, .constant = 5 },
	{ "NL", ZP_UNKNOWN }, /* Newfoundland is CQ zone 5, Labrador is 2 */
	{ "YT", ZP_CONST, .constant = 1 },
	{ "PE", ZP_CONST, .constant = 5 },
	/*
	 * This isn't really a straight line.  There are various islands
	 * that are semi-arbitrarily assigned between zone 1 and 2 instead
	 * of just a simple longitude cutoff.  See zone 1 & 2 descriptions.
	 */
	{ "NU", ZP_WESTOF,
		.latlon.boundary = -102.,
		.latlon.zones = { 2, 1 } },
};

static const struct zone usa[] = {
	/* call area 1 */
	{ "CT", ZP_CONST, .constant = 5 },
	{ "ME", ZP_CONST, .constant = 5 },
	{ "MA", ZP_CONST, .constant = 5 },
	{ "NH", ZP_CONST, .constant = 5 },
	{ "RI", ZP_CONST, .constant = 5 },
	{ "VT", ZP_CONST, .constant = 5 },

	/* call area 2 */
	{ "NJ", ZP_CONST, .constant = 5 },
	{ "NY", ZP_CONST, .constant = 5 },

	/* call area 3 */
	{ "DC", ZP_CONST, .constant = 5 }, /* not really a state */
	{ "DE", ZP_CONST, .constant = 5 },
	{ "MD", ZP_CONST, .constant = 5 },
	{ "PA", ZP_CONST, .constant = 5 },

	/* call area 4 */
	{ "AL", ZP_CONST, .constant = 4 },
	{ "FL", ZP_CONST, .constant = 5 },
	{ "GA", ZP_CONST, .constant = 5 },
	{ "KY", ZP_CONST, .constant = 4 },
	{ "NC", ZP_CONST, .constant = 5 },
	{ "SC", ZP_CONST, .constant = 5 },
	{ "TN", ZP_CONST, .constant = 4 },
	{ "VA", ZP_CONST, .constant = 5 },

	/* call area 5 */
	{ "AR", ZP_CONST, .constant = 4 },
	{ "LA", ZP_CONST, .constant = 4 },
	{ "MS", ZP_CONST, .constant = 4 },
	{ "NM", ZP_CONST, .constant = 4 },
	{ "OK", ZP_CONST, .constant = 4 },
	{ "TX", ZP_CONST, .constant = 4 },

	/* call area 6 */
	{ "CA", ZP_CONST, .constant = 3 },

	/* call area 7 */
	{ "AZ", ZP_CONST, .constant = 3 },
	{ "ID", ZP_CONST, .constant = 3 },
	{ "MT", ZP_CONST, .constant = 4 },
	{ "NV", ZP_CONST, .constant = 3 },
	{ "OR", ZP_CONST, .constant = 3 },
	{ "UT", ZP_CONST, .constant = 3 },
	{ "WA", ZP_CONST, .constant = 3 },
	{ "WY", ZP_CONST, .constant = 4 },

	/* call area 8 */
	{ "MI", ZP_CONST, .constant = 4 },
	{ "OH", ZP_CONST, .constant = 4 },
	{ "WV", ZP_CONST, .constant = 5 },

	/* call area 9 */
	{ "IL", ZP_CONST, .constant = 4 },
	{ "IN", ZP_CONST, .constant = 4 },
	{ "WI", ZP_CONST, .constant = 4 },

	/* call area 0 */
	{ "CO", ZP_CONST, .constant = 4 },
	{ "IA", ZP_CONST, .constant = 4 },
	{ "KS", ZP_CONST, .constant = 4 },
	{ "MN", ZP_CONST, .constant = 4 },
	{ "MO", ZP_CONST, .constant = 4 },
	{ "NE", ZP_CONST, .constant = 4 },
	{ "ND", ZP_CONST, .constant = 4 },
	{ "SD", ZP_CONST, .constant = 4 },
};

int cqz_by_state(int dxcc, const char *state)
{
	switch (dxcc) {
		case 1: /* Canada */
			return find_zone(state, canada, ARRAY_LEN(canada),
					 false, 0, false, 0);
		case 6: /* Alaska */
			return 1;
		case 110: /* Hawaii */
			return 31;
		case 291: /* United States */
			return find_zone(state, usa, ARRAY_LEN(usa),
					 false, 0, false, 0);
	}

	return -1;
}
