/*
 * Copyright (c) 2021-2022,2024 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

/*
 * Unfortunately, there doesn't seem to be an exact definition of ITU zones.
 * In the USA, the ARRL's definition is what everyone uses as the official
 * one.
 *
 * https://www.arrl.org/iaru-hf-world-championship
 *
 * As of 2024-05-11, the contents were:
 *
 * -------------------------------------8<-------------------------------------
 *  ITU Zone    Prefix    Location
 *
 *  06          W-K-N-A   U.S.A. (Washington, Oregon, California, Nevada,
 *			  Idaho, & that part of Montana, Utah & Arizona west
 *			  of 110 Deg. W.)
 *
 *  07          W-K-N-A   U.S.A. (N. Dakota, S. Dakota, Nebraska, Wyoming,
 *			  Colorado, New Mexico, Texas, Oklahoma, Kansas,
 *			  Iowa, Minnesota, that part of Utah & Arizona East
 *			  of 110 Deg. W., & that part of Michigan, part of
 *			  Montana Illinois, Missouri, Arkansas, Tennessee,
 *			  Mississippi, Louisiana & Wisconsin west of 90 Deg.
 *			  W.)
 *
 *  08          W-K-N-A   U.S.A. (Indiana, Alabama, Georgia, Florida,
 *			  Virginia, Kentucky, N. Carolina, S.  Carolina,
 *			  West Virginia, Maryland, Delaware, Ohio,
 *			  Pennsylvania, New Jersey, New York, Maine,
 *			  Connecticut, Rhode Island, Massachusetts, New
 *			  Hampshire, Vermont, & that part of Michigan,
 *			  Illinois, Missouri, Arkansas, Mississippi,
 *			  Tennessee, Louisiana & Wisconsin east of 90 Deg.
 *			  W.)
 *
 *  01          KL7       Alaska (West of 141 Deg. W.)
 *
 *  02          KL7       Alaska (East of 141 Deg. W.)
 *              VE6,7,VY1 Canada (South of 80 Deg. N. and West of 110 Deg.
 *	                  W.)
 *
 *  03          VE3,4,5,8 Canada (South of 80 Deg. N. and between 90 Deg.
 *			  and 110 Deg. W.) Includes western area of Ontario
 *			  North section
 *
 *  04          Pieces    Canada (South of 80 Deg. N. and
 *              of VY0    between 70 Deg. and 90 Deg. W.
 *              VE2,3,4,8 including Baffin Isl.)
 *                        Includes Ontario East, Ontario South & Greater
 *                        Toronto Area sections and eastern part of
 *                        Ontario North section
 *
 *  09          FP        St. Pierre et Miquelon
 *              VE1,2
 *              VY2,VO2,  Canada (South of 80 Deg. N. &
 *              VO1       East of 70 Deg. W., including
 *              Pieces    Labrador, Newfoundland, Nova
 *              of VY0    Scotia, but excluding Baffin Isl.)
 *  75          VE8       Canada (80 Deg. N. to 90 Deg. N., North Pole)
 * -------------------------------------8<-------------------------------------
 */

#include "zones_impl.h"

static const struct zone canada[13] = {
	{ "NS", ZP_CONST, .constant = 9 },
	{ "QC", ZP_UNKNOWN }, /* 4 or 9 */
	{ "ON", ZP_UNKNOWN }, /* 3 or 4 */
	{ "MB", ZP_CONST, .constant = 3 },
	{ "SK", ZP_CONST, .constant = 3 },
	{ "AB", ZP_CONST, .constant = 2 },
	{ "BC", ZP_CONST, .constant = 2 },
	{ "NT", ZP_UNKNOWN }, /* 2 or 3 */
	{ "NB", ZP_CONST, .constant = 9 },
	{ "NL", ZP_CONST, .constant = 9 },
	{ "YT", ZP_CONST, .constant = 2 },
	{ "PE", ZP_CONST, .constant = 9 },
	{ "NU", ZP_UNKNOWN }, /* 3 or 4 */
};

static const struct zone usa[] = {
	/* call area 1 */
	{ "CT", ZP_CONST, .constant = 8 },
	{ "ME", ZP_CONST, .constant = 8 },
	{ "MA", ZP_CONST, .constant = 8 },
	{ "NH", ZP_CONST, .constant = 8 },
	{ "RI", ZP_CONST, .constant = 8 },
	{ "VT", ZP_CONST, .constant = 8 },

	/* call area 2 */
	{ "NJ", ZP_CONST, .constant = 8 },
	{ "NY", ZP_CONST, .constant = 8 },

	/* call area 3 */
	{ "DC", ZP_CONST, .constant = 8 }, /* not really a state */
	{ "DE", ZP_CONST, .constant = 8 },
	{ "MD", ZP_CONST, .constant = 8 },
	{ "PA", ZP_CONST, .constant = 8 },

	/* call area 4 */
	{ "AL", ZP_CONST, .constant = 8 },
	{ "FL", ZP_CONST, .constant = 8 },
	{ "GA", ZP_CONST, .constant = 8 },
	{ "KY", ZP_CONST, .constant = 8 },
	{ "NC", ZP_CONST, .constant = 8 },
	{ "SC", ZP_CONST, .constant = 8 },
	{ "TN", ZP_UNKNOWN }, /* 7 or 8 (lon >= -90) */
	{ "VA", ZP_CONST, .constant = 8 },

	/* call area 5 */
	{ "AR", ZP_UNKNOWN }, /* 7 or 8 (lon > -90) */
	{ "LA", ZP_UNKNOWN }, /* 7 or 8 (lon > -90) */
	{ "MS", ZP_UNKNOWN }, /* 7 or 8 (lon > -90) */
	{ "NM", ZP_CONST, .constant = 7 },
	{ "OK", ZP_CONST, .constant = 7 },
	{ "TX", ZP_CONST, .constant = 7 },

	/* call area 6 */
	{ "CA", ZP_CONST, .constant = 6 },

	/* call area 7 */
	{ "AZ", ZP_UNKNOWN }, /* 6 or 7 (lon > -110) */
	{ "ID", ZP_CONST, .constant = 6 },
	{ "MT", ZP_UNKNOWN }, /* 6 or 7 (lon > -110) */
	{ "NV", ZP_CONST, .constant = 6 },
	{ "OR", ZP_CONST, .constant = 6 },
	{ "UT", ZP_UNKNOWN }, /* 6 or 7 (lon > -110) */
	{ "WA", ZP_CONST, .constant = 6 },
	{ "WY", ZP_CONST, .constant = 7 },

	/* call area 8 */
	{ "MI", ZP_UNKNOWN }, /* 7 or 8 (lon >= -90) */
	{ "OH", ZP_CONST, .constant = 8 },
	{ "WV", ZP_CONST, .constant = 8 },

	/* call area 9 */
	{ "IL", ZP_UNKNOWN }, /* 7 or 8 (lon >= -90) */
	{ "IN", ZP_CONST, .constant = 8 },
	{ "WI", ZP_UNKNOWN }, /* 7 or 8 (lon >= -90) */

	/* call area 0 */
	{ "CO", ZP_CONST, .constant = 7 },
	{ "IA", ZP_CONST, .constant = 7 },
	{ "KS", ZP_CONST, .constant = 7 },
	{ "MN", ZP_UNKNOWN }, /* 7 or 8 (lon >= -90) */
	{ "MO", ZP_UNKNOWN }, /* 7 or 8 (lon >= -90) */
	{ "NE", ZP_CONST, .constant = 7 },
	{ "ND", ZP_CONST, .constant = 7 },
	{ "SD", ZP_CONST, .constant = 7 },
};

int ituz_by_state(int dxcc, const char *state)
{
	switch (dxcc) {
		case 1: /* Canada */
			return find_zone(state, canada, ARRAY_LEN(canada),
					 false, 0, false, 0);
		case 6: /* Alaska */
			return -1; /* 1 or 2 */
		case 110: /* Hawaii */
			return 61;
		case 291: /* United States */
			return find_zone(state, usa, ARRAY_LEN(usa),
					 false, 0, false, 0);
	}

	return -1;
}
