/*
 * Copyright (c) 2021-2023 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/version.h>
#include <jeffpc/error.h>
#include <jeffpc/io.h>
#include <jeffpc/val.h>
#include <jeffpc/rbtree.h>

#include <hlog/version.h>
#include <hlog/qso.h>
#include <hlog/qso-pack.h>
#include <hlog/index.h>
#include <hlog/util.h>

#define NEW_FILE_FNAME	"create-new"

static struct qso *load(const char *fname, bool create_new)
{
	struct qso *qso;
	size_t len;
	char *raw;
	int ret;

	qso = qso_alloc();
	if (!qso)
		return ERR_PTR(-ENOMEM);

	if (create_new)
		return qso; /* will use auto-generated uuid */

	raw = read_file_len(fname, &len);
	if (IS_ERR(raw)) {
		ret = PTR_ERR(raw);
		goto err;
	}

	ret = qso_unpack_hlq(qso, raw, len);
	if (ret)
		goto err_raw;

	free(raw);

	return qso;

err_raw:
	free(raw);

err:
	qso_putref(qso);

	return ERR_PTR(ret);
}

static void usage(const char *prog)
{
	fprintf(stderr, "HLog version %s (libjeffpc %s)\n\n", version_string,
		jeffpc_version);

	fprintf(stderr, "Usage: %s <qso filename> [hlq key-type-value] ...\n",
		prog);
	fprintf(stderr, "\n");
	fprintf(stderr, "If the filename is '%s', a new file is created "
		"automatically.\n", NEW_FILE_FNAME);
	exit(1);
}

int main(int argc, char **argv)
{
	const char *qsofile;
	bool create_new;
	struct qso *qso;
	int ret;
	int i;

	if (argc < 2)
		usage(argv[0]);

	qsofile = argv[1];
	create_new = !strcmp(qsofile, NEW_FILE_FNAME);

	qso = load(qsofile, create_new);
	if (IS_ERR(qso)) {
		fprintf(stderr, "Error: failed to open QSO file: %s\n",
			xstrerror(PTR_ERR(qso)));
		goto err;
	}

	for (i = 2; i < argc; i++) {
		const char *key_start;
		const char *key_end;
		struct buffer buf;
		struct val *val;
		enum qso_ns ns;
		size_t kvlen;
		char *kv;

		/* add a '\n' since that's what the parser expects */
		kvlen = strlen(argv[i]);
		kv = alloca(kvlen + 2);
		memcpy(kv, argv[i], kvlen);
		kv[kvlen] = '\n';
		kv[kvlen + 1] = '\0';

		buffer_init_static(&buf, kv, strlen(kv), strlen(kv), false);

		val = hlq_parse_kv(&buf, HLQ_VERSION_USE_DEFAULT, &ns,
				   &key_start, &key_end);
		if (IS_ERR(val)) {
			fprintf(stderr, "Error: failed to parse item (%s): %s\n",
				argv[i], xstrerror(PTR_ERR(val)));
			goto err_free;
		}

		ret = qso_set_hlq_item(qso, HLQ_VERSION_USE_DEFAULT, ns,
				       key_start, key_end, val);
		if (ret) {
			fprintf(stderr, "Error: failed to set item (%s): %s\n",
				argv[i], xstrerror(ret));
			goto err_free;
		}
	}

	ret = qso_save(qso, AT_FDCWD);
	if (ret) {
		fprintf(stderr, "Error: failed to save QSO: %s\n",
			xstrerror(ret));
		goto err_free;
	}

	printf("%s\n", qso->uuid_str);

	qso_putref(qso);

	return 0;

err_free:
	qso_putref(qso);

err:
	return 1;
}
