/*
 * Copyright (c) 2020-2022 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/io.h>
#include <jeffpc/time.h>

#include <hlog/country.h>
#include <hlog/sig.h>

#include "databases.h"

#define DATA_DIR_COUNTRY_FNAME	"data/cty.csv"
#define DATA_DIR_POTA_FNAME	"data/pota_parks.csv"
#define DATA_DIR_SOTA_FNAME	"data/sota_summits.csv"

static int load_countries(void)
{
	int ret;

	printf("Loading countries data from %s...\n", DATA_DIR_COUNTRY_FNAME);
	ret = country_load(AT_FDCWD, DATA_DIR_COUNTRY_FNAME);
	printf("%s.\n", ret ? "error" : "done");

	if (ret)
		fprintf(stderr, "Error: failed to load country data: %s\n",
			xstrerror(ret));

	return ret;
}

static void load_sig(const char *name, const char *fname,
		     int (*f)(int, const char *))
{
	uint64_t start,end;
	int ret;

	printf("Loading %s reference data from %s...\n", name, fname);

	start = gettime();
	ret = f(AT_FDCWD, fname);
	end = gettime();

	if (ret)
		printf("Warning: Failed to load %s reference data: %s\n", name,
			xstrerror(ret));
	else
		printf("Loaded %s data in %"PRIu64".%03"PRIu64" secs\n", name,
			(end - start) / 1000000000,
			(end - start) % 1000000000 / 1000000);
}

int load_databases(void)
{
	int ret;

	/* first, load the required ones */
	ret = load_countries();
	if (ret)
		return ret;

	/* second, load any optional ones */
	load_sig("POTA", DATA_DIR_POTA_FNAME, sig_load_pota);
	load_sig("SOTA", DATA_DIR_SOTA_FNAME, sig_load_sota);

	return 0;
}
