/*
 * Copyright (c) 2020-2022 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/mem.h>
#include <jeffpc/unicode.h>

#include <hlog/index.h>
#include <hlog/qso_line.h>

#include "recent.h"
#include "xcurses.h"

struct recent_line {
	struct xuuid uuid;
	char str[];
};

int recent_init(struct recent *r, int height, int width, int row, int col,
		bool all)
{
	size_t i;

	r->win = newwin(height, width, row, col);
	whrule(r->win, height - 1, width);

	r->nrows = height;
	r->ncols = width;
	r->lines = mem_reallocarray(NULL, r->nrows, sizeof(struct recent_line));
	if (!r->lines)
		goto err;

	for (i = 0; i < r->nrows; i++) {
		r->lines[i] = zalloc(sizeof(struct recent_line) + r->ncols + 1);
		if (!r->lines[i])
			goto err_free;
	}

	r->all = all;
	r->reeval = true;

	return 0;

err_free:
	for (i = 0; i < r->nrows; i++)
		free(r->lines[i]);

	free(r->lines);

err:
	delwin(r->win);

	return -ENOMEM;
}

void recent_destroy(struct recent *r)
{
	size_t i;

	for (i = 0; i < r->nrows; i++)
		free(r->lines[i]);

	free(r->lines);

	delwin(r->win);
}

void recent_refresh(struct recent *r, struct str *call)
{
	struct qso *(*next)(struct index_iter_ctx *);
	struct index_iter_ctx ctx;
	struct qso *qso;
	size_t i;

	if (!r->reeval) {
		str_putref(call);
		return;
	}

	if (call) {
		/* per-call interation with a call specified */
		ASSERT(!r->all);
		qso = index_iter_call_history_start(call, &ctx, false);
		next = index_iter_call_history_next;

		str_putref(call);
	} else if (!r->all) {
		/* per-call iteration but without a call specified */
		qso = NULL;
		next = NULL;
	} else {
		/* all-history iteration */
		qso = index_iter_history_start(&ctx, false);
		next = index_iter_history_next;
	}

	/* cache the text */
	for (i = 0;
	     qso && (i < r->nrows);
	     qso = next(&ctx), i++) {
		r->lines[i]->uuid = qso->uuid;
		qso_line_fmt(qso, r->lines[i]->str, r->ncols);
	}

	/* clear the remaining lines */
	for (; i < r->nrows; i++)
		memset(r->lines[i]->str, ' ', r->ncols);

	/* print cached text */
	for (i = 0; i < r->nrows; i++)
		mvwprintw(r->win, i, 0, "%-*s", r->ncols, r->lines[i]->str);

	wrefresh(r->win);

	r->reeval = false;
}

void recent_invalidate(struct recent *r)
{
	r->reeval = true;
}

/*
 * Legacy recent code, to be removed in the future
 */
void refresh_recent_win(WINDOW *win, int height, int width, bool count)
{
	const int start_row = count ? 1 : 0;
	struct index_iter_ctx ctx;
	size_t total_qsos;
	struct qso *qso;
	int row;

	/* clear it */
	for (row = start_row; row < height; row++)
	     mvwprintw(win, row, 0, "%*s", width - 1, "");

	/* for each mode, print a line */
	for (qso = index_iter_history_start(&ctx, false), row = start_row;
	     qso && (row < height);
	     qso = index_iter_history_next(&ctx), row++) {
		char tmp[width];

		wmove(win, row, 0);

		qso_line_fmt(qso, tmp, sizeof(tmp));
		wprintw(win, "%s", tmp);
	}

	if (!count)
		return;

	/* print QSO count in header line */
	index_counts(&total_qsos);
	if (total_qsos) {
		char tmp[width];

		if (total_qsos > 1)
			snprintf(tmp, sizeof(tmp), "1…%d/%zu", row - 1,
				 total_qsos);
		else
			snprintf(tmp, sizeof(tmp), "1/1");

		/*
		 * HACK: count number of printable characters
		 */
		wmove(win, 0, width -
		      utf8_is_valid_str(tmp, strlen(tmp)) - 1);
		wprintw(win, "%s", tmp);
	}
}
