/*
 * Copyright (c) 2021-2022 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#ifndef __XFORM_H
#define __XFORM_H

#include <ncurses.h>

#include <jeffpc/str.h>

enum xform_req {
	XFORM_REQ_FIRST_FIELD = -100,
	XFORM_REQ_NEXT_FIELD, /* got to the next field */
	XFORM_REQ_BEG_FIELD, /* go to the start of the current field */
	XFORM_REQ_END_FIELD, /* go to the end of the current field */
	XFORM_REQ_PREV_CHAR, /* got back one char in the current field */
	XFORM_REQ_NEXT_CHAR, /* got forward one char in the current field */
	XFORM_REQ_DEL_PREV, /* delete previous char in the current field */
	XFORM_REQ_DEL_THIS, /* delete current char in the current field */
};

enum xform_color {
	XFORM_COLOR_NONE,
	XFORM_COLOR_BLACK,
	XFORM_COLOR_RED,
	XFORM_COLOR_GREEN,
	XFORM_COLOR_YELLOW,
	XFORM_COLOR_BLUE,
	XFORM_COLOR_MAGENTA,
	XFORM_COLOR_CYAN,
	XFORM_COLOR_WHITE,
};
#define XFORM_NUM_COLORS 9

struct xform_palette {
	int pair[XFORM_NUM_COLORS];
};

enum xform_underline {
	XFORM_UNDERLINE_AUTO,
	XFORM_UNDERLINE_OFF,
	XFORM_UNDERLINE_ON,
};

struct xfield;

struct xfield_def {
	enum {
		XFIELD_TYPE_LABEL,
		XFIELD_TYPE_FIELD,
		XFIELD_TYPE_ANNUNCIATOR,
	} type;

	/* location */
	int row;
	int col;
	int width;

	/* label */
	struct str *label;

	/* attributes */
	struct {
		enum xform_color color;
		enum xform_underline underline;
	} attr;

	void *private;
};

struct xform {
	WINDOW *basewin;
	WINDOW *formwin;
	struct xform_palette palette;
	struct xfield *fields;
	struct xfield *labels;
	struct xfield *annunc;
	size_t nfields;
	size_t nlabels;
	size_t nannunc;

	size_t curfield;
};

/*
 * Form
 */
extern int xform_alloc(struct xform *form, int height, int width, int row,
		       int col, const struct xform_palette *palette,
		       const struct xfield_def *fields, size_t nfields);
extern void xform_free(struct xform *form);
extern int xform_driver(struct xform *form, int req);

/* return entered/private data */
extern const void *xform_field_get(struct xform *form, int idx, size_t *size_r);
extern void *xform_field_get_private(struct xform *form, int idx);
/* set entered/private data */
extern void xform_field_set(struct xform *form, int idx, const char *data);
extern void xform_field_set_private(struct xform *form, int idx, void *private);
/* clear the entered data */
extern void xform_field_clear(struct xform *form, int idx);

/* get labels */
extern struct str *xform_annunciator_get_label(struct xform *form, int idx);

/* set annunciator color */
extern void xform_annunciator_set(struct xform *form, int idx,
				  enum xform_color color);
/* reset annunciator color (use color set during initialization) */
extern void xform_annunciator_clear(struct xform *form, int idx);

extern void xform_cursor_warp(struct xform *form);

/* conversion functions */
enum xform_color xform_color_from_name(const char *str);

static inline void xform_refresh(struct xform *form)
{
	wrefresh(form->basewin);
	wrefresh(form->formwin);
}

static inline int xform_getch(struct xform *form)
{
	return wgetch(form->formwin);
}

#endif
