/*
 * Copyright (c) 2020-2024 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/version.h>
#include <jeffpc/error.h>
#include <jeffpc/time.h>
#include <jeffpc/val.h>
#include <jeffpc/io.h>
#include <jeffpc/mem.h>

#include <ctype.h>
#include <unistd.h>
#include <locale.h>
#include <math.h>
#include <spawn.h>

#include <hlog/version.h>
#include <hlog/qso.h>
#include <hlog/qso_line.h>
#include <hlog/index.h>
#include <hlog/util.h>
#include <hlog-lua/xlua.h>

#include <hlog-rpc/announce-pub.h>

#include "common/databases.h"
#include "common/xcurses.h"
#include "common/xform.h"

#include "contest/contest.h"
#include "contest/location.h"
#include "contest/rig.h"
#include "contest/script.h"
#include "contest/wsjtx.h"

#define DEFAULT_RIG_HOST	"localhost"
#define DEFAULT_RIG_PORT	4532

static void usage(const char *prog, const char *fmt, ...)
{
	if (fmt) {
		va_list ap;

		fprintf(stderr, "Error: ");

		va_start(ap, fmt);
		vfprintf(stderr, fmt, ap);
		va_end(ap);

		fprintf(stderr, "\n");
	}

#define P(...) fprintf(stderr, __VA_ARGS__)

	P("hlog version %s (libjeffpc %s)\n\n", version_string, jeffpc_version);

	P("Usage: %s -c <contest> [-y <year>] [-E <path>]\n", prog);
	P("          [-R <hostname>] [-p <port>] [-m <model>] [-r <device>]\n");
	P("          <data dir>\n");
	P("\n");
	P("Options:\n");
	P("  -E <path>      - export contest log to path\n");
	P("  -c <contest>   - name of contest (e.g., CQ-WW-SSB)\n");
	P("  -y <year>      - year of contest (defaults to current year)\n");
	P("  -R <hostname>  - Flex rig hostname to listen on (default: %s)\n",
	  DEFAULT_RIG_HOST);
	P("  -p <port>      - Flex rig TCP port to listen on (default: %u)\n",
	  DEFAULT_RIG_PORT);
	P("  -m <model>     - hamlib rig model\n");
	P("  -r <device>    - hamlib rig device\n");
	P("\n");
	P("If neither the hamlib device nor the model are not specified, rig\n"
	  "control will not be available.\n");
	P("\n");
	P("If neither Flex rig hostname nor port are specified, the "
	  "defaults are ignored\n");
	P("and rig control services are not provided.\n");

	exit(1);

#undef P
}

static int load_index(void)
{
	int ret;

	printf("Loading existing QSOs...\n");
	ret = index_init(AT_FDCWD);
	printf("%s.\n", ret ? "error" : "done");

	if (ret)
		fprintf(stderr, "Error: failed to load existing entries: %s\n",
			xstrerror(ret));

	return ret;
}

static int start_listeners(uint32_t rig_model, const char *rig_dev,
			   const char *rig_host, uint16_t rig_port)
{
	int ret;

	ret = announce_start();
	if (ret)
		return ret;

	ret = rig_start(rig_model, rig_dev, rig_host, rig_port);
	if (ret)
		goto err;

	ret = location_start();
	if (ret)
		goto err_rig;

	ret = wsjtx_start();
	if (ret)
		goto err_loc;

	return 0;

err_loc:
	location_stop();

err_rig:
	rig_stop();

err:
	announce_stop();

	return ret;
}

static void stop_listeners(void)
{
	wsjtx_stop();
	location_stop();
	rig_stop();
	announce_stop();
}

static int do_export(lua_State *L, const struct contest_params *params,
		     const char *outdir)
{
	char error[128];
	int ret;

	printf("Exporting %s entries to %s...\n", params->name, outdir);

	ret = script_event_export(L, outdir, error, sizeof(error));
	if (ret)
		fprintf(stderr, "Error: Failed to export log: %s\n",
			error);

	return ret;
}

int main(int argc, char **argv)
{
	struct xfield_def *wfields = NULL;
	size_t nwfields = 0;
	struct contest_params contest_params = { NULL, 0 };
	struct qso *template;
	const char *data_dir;
	const char *rig_host;
	const char *rig_dev;
	uint32_t rig_model;
	uint16_t rig_port;
	lua_State *L;
	char *export;
	int ret;
	int opt;

	setlocale(LC_ALL, "");
	ASSERT0(putenv("TZ=UTC"));

	export = NULL;
	rig_host = NULL;
	rig_dev = NULL;
	rig_model = 0;
	rig_port = 0;
	while ((opt = getopt(argc, argv, "E:c:y:R:p:m:r:")) != -1) {
		switch (opt) {
			case 'E':
				export = realpath(optarg, NULL);
				if (!export)
					usage(argv[0], "invalid export path "
					      "specified: %s", strerror(errno));
				break;
			case 'c':
				contest_params.name = optarg;
				break;
			case 'y':
				ret = str2u32(optarg, &contest_params.year);
				if (ret || (contest_params.year < 1900) ||
				    (contest_params.year > 2200))
					usage(argv[0], "invalid year '%s'",
					      optarg);
				break;
			case 'R':
				if (rig_host)
					usage(argv[0], "hamlib rig hostname "
					      "specified multiple times");

				rig_host = optarg;
				break;
			case 'p':
				if (rig_port)
					usage(argv[0], "hamlib rig port "
					      "specified multiple times");

				ret = str2u16(optarg, &rig_port);
				if (ret)
					usage(argv[0], "invalid hamlib rig "
					      "port: %s", xstrerror(ret));

				if (!rig_port)
					usage(argv[0], "invalid hamlib rig "
					      "port: zero port number");
				break;
			case 'm':
				if (rig_model)
					usage(argv[0], "hamlib rig model "
					      "specified multiple times");

				ret = str2u32(optarg, &rig_model);
				if (ret)
					usage(argv[0], "invalid hamlib rig "
					      "model: %s", xstrerror(ret));

				if (!rig_model)
					usage(argv[0], "invalid hamlib rig "
					      "model: zero model number");
				break;
			case 'r':
				if (rig_dev)
					usage(argv[0], "hamlib rig device "
					      "specified multiple times");

				rig_dev = optarg;
				break;
			default:
				usage(argv[0], NULL);
				break;
		}
	}

	if (!contest_params.name)
		usage(argv[0], "contest name required");

	/*
	 * make sure either both rig model and device are specified, or
	 * neither is specified
	 */
	if ((rig_model || rig_dev) && !(rig_model && rig_dev))
		usage(argv[0], "rig model & device must be both specified");

	/*
	 * make sure both rig host and port are both set if at least one of
	 * them is specified
	 */
	if (rig_host || rig_port) {
		rig_host = rig_host ? rig_host : DEFAULT_RIG_HOST;
		rig_port = rig_port ? rig_port : DEFAULT_RIG_PORT;
	}

	ret = script_exists(&contest_params);
	if (ret)
		usage(argv[0], "unknown contest '%s'", contest_params.name);

	/* missing data dir path */
	if ((optind + 1) != argc)
		usage(argv[0], "path to data directory required");

	data_dir = argv[argc - 1];

	/* contest year defaults to "now" */
	if (!contest_params.year) {
		struct tm tm;
		time_t now;

		now = time(NULL);
		gmtime_r(&now, &tm);

		contest_params.year = tm.tm_year + 1900;
	}

	/*
	 * chdir to the data directory to make our lives in the future much
	 * easier.  We will be able to use relative paths everywhere.
	 */
	if (chdir(data_dir)) {
		fprintf(stderr, "Error: failed to change to data dir: %s\n",
			strerror(errno));
		ret = -errno;
		goto err;
	}

	ret = load_databases();
	if (ret)
		goto err;

	ret = load_index();
	if (ret)
		goto err;

	if (!export) {
		ret = msg_init();
		if (ret)
			goto err;

		ret = start_listeners(rig_model, rig_dev, rig_host, rig_port);
		if (ret)
			goto err_msg;

		ret = contest_init(&contest_params, &template);
		if (ret)
			goto err_listen;
	} else {
		template = NULL;
	}

	L = script_load(&wfields, &nwfields, &contest_params, template);
	if (IS_ERR(L)) {
		ret = PTR_ERR(L);
		goto err_free;
	}

	if (!export)
		ret = contest_run(L, wfields, nwfields);
	else
		ret = do_export(L, &contest_params, export);

	lua_close(L);

err_free:
	if (!export)
		contest_cleanup();

err_listen:
	if (!export)
		stop_listeners();

err_msg:
	if (!export)
		msg_free();

err:
	free(export);

	return ret ? 1 : 0;
}
