/*
 * Copyright (c) 2022,2024 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/jeffpc.h>
#include <jeffpc/synch.h>
#include <jeffpc/mem.h>

#include "contest.h"

#define NUM_MSGS		10000 /* keep 10k messages */
#define TYPICAL_MSG_LEN		100 /* bytes */
#define BUF_SIZE		(NUM_MSGS * TYPICAL_MSG_LEN)

struct msg {
	enum msg_src src:8;
	enum errlevel level:8;
	uint8_t len;
	struct time ts;
	bool trunc:1;
	bool inuse:1;
	char data[TYPICAL_MSG_LEN];
};

static LOCK_CLASS(msg_lc);

static struct lock lock;
static struct msg *msgs;
static size_t next_msg; /* index for next message */

void msg_get_msgs(bool (*cb)(enum msg_src src, enum errlevel level,
			     const struct time *ts, const char *msg,
			     void *private),
		  void *private)
{
	size_t i;

	MXLOCK(&lock);

	/* scan backwards until we find a !inuse message */
	for (i = 0; i < NUM_MSGS; i++) {
		size_t idx;

		if (i < next_msg)
			idx = next_msg - i - 1;
		else
			idx = NUM_MSGS - i + next_msg - 1;

		if (!msgs[idx].inuse)
			break;

		if (!cb(msgs[idx].src, msgs[idx].level, &msgs[idx].ts,
			msgs[idx].data, private))
			break;
	}

	MXUNLOCK(&lock);
}

static void msg_append(const struct time *now, enum msg_src src,
		       enum errlevel level, const char *start,
		       size_t ideal_len)
{
	struct msg new;
	size_t len;

	if (!ideal_len)
		return; /* ignore empty messages */

	len = MIN(ideal_len, sizeof(new.data) - 1);

	memcpy(new.data, start, len);
	new.data[len] = '\0';

	new.src = src;
	new.level = level;
	new.len = MIN(ideal_len, UINT8_MAX);
	new.ts = *now;
	new.trunc = ideal_len != new.len;
	new.inuse = true;

	/* TODO: write the message to a log file */

	msgs[next_msg] = new;
	next_msg = (next_msg + 1) % NUM_MSGS;
}

static void msg_vprint(enum msg_src src, enum errlevel level, const char *fmt,
		       va_list ap)
{
	struct time now;
	char buf[4096];
	char *start;
	char *end;

	time_set_now(&now);

	vsnprintf(buf, sizeof(buf), fmt, ap);

	MXLOCK(&lock);

	for (start = buf; (end = strchr(start, '\n')); start = end + 1)
		msg_append(&now, src, level, start, end - start);

	msg_append(&now, src, level, start, strlen(start));

	MXUNLOCK(&lock);
}

static void msg_sys_vprint(enum errlevel level, const char *fmt, va_list ap)
{
	msg_vprint(MSG_SYS, level, fmt, ap);
}

void msg_print(enum msg_src src, enum errlevel level, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	msg_vprint(src, level, fmt, ap);
	va_end(ap);
}

int msg_init(void)
{
	msgs = zalloc(sizeof(struct msg) * NUM_MSGS);
	if (!msgs)
		return -ENOMEM;

	MXINIT(&lock, &msg_lc);

	next_msg = 0;

	return 0;
}

void msg_start(void)
{
	struct jeffpc_ops ops = {
		.print = msg_sys_vprint,
	};

	/* redirect cmn_err to us */
	jeffpc_init(&ops);
}

void msg_stop(void)
{
	jeffpc_init(NULL);

	/* TODO: dump out messages */
}

void msg_free(void)
{
	MXDESTROY(&lock);

	free(msgs);
}
