/*
 * Copyright (c) 2021-2024 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <unistd.h>

#include <jeffpc/types.h>
#include <jeffpc/error.h>
#include <jeffpc/thread.h>
#include <jeffpc/time.h>
#include <jeffpc/socksvc.h>

#include <hlog-rpc/announce.h>
#include <hlog-rpc/announce-pub.h>
#include <hlog-rpc/hamlib.h>

#include "rig.h"
#include "rigctl.h"

/* How long to wait between reconnect attempts */
#define RECONNECT_DELAY 10 /* s */

/*
 * How long to wait between ANNOUNCE_RIG_FREQ notifications even if there
 * were no changes
 */
#define UPDATE_TIMEOUT	(60ull * 1000 * 1000 * 1000) /* ns */

static uint32_t rig_model;
static const char *rig_dev;
static struct hrig *rig;
static pthread_t rig_thread;
static pthread_t rigctl_thread;

static void notify(enum announce_rig_msg_type type, uint64_t tx_freq,
		   uint64_t rx_freq, enum hrig_mode tx_mode,
		   enum hrig_mode rx_mode, uint32_t power)
{
	static enum hrig_mode ptxm, prxm;
	static uint64_t last_update;
	static uint64_t ptx, prx;
	static uint32_t ppower;
	bool changed;

	changed = (ptx != tx_freq) || (prx != rx_freq) || (ppower != power) ||
		(ptxm != tx_mode) || (prxm != rx_mode);

	switch (type) {
		case ANNOUNCE_RIG_FREQ:
			if (!changed &&
			    (last_update + UPDATE_TIMEOUT) > gettime())
				return;
			break;
		case ANNOUNCE_RIG_DISCONNECT:
			tx_freq = 0;
			rx_freq = 0;
			power = 0;
			tx_mode = HRIG_MODE_UNKNOWN;
			rx_mode = HRIG_MODE_UNKNOWN;
			break;
	}

	ptx = tx_freq;
	prx = rx_freq;
	ppower = power;
	ptxm = tx_mode;
	prxm = rx_mode;
	last_update = gettime();

	announce_publish_rig(type, tx_freq, rx_freq, tx_mode, rx_mode, power);
}

static bool rig_connect(void)
{
	/*
	 * To avoid spamming the log with reconnection attempts, we only
	 * print the error messages only once until we get a connection.
	 * Then, if we lose the connection, we print the messages once.  And
	 * so on.
	 */
	static bool silence;
	int ret;

	ret = hamlib_open(rig);
	switch (ret) {
		case 0:
			cmn_err(CE_INFO, "Connected to %s", rig_dev);
			silence = false;
			return true;
		case -EBUSY:
			/* already connected */
			return true;
		default:
			if (!silence)
				cmn_err(CE_INFO, "Failed to open rig: %s",
					xstrerror(ret));
			silence = true;
			return false;
	}
}

static void rig_disconnect(void)
{
	int ret;

	ret = hamlib_close(rig);
	if (!ret)
		cmn_err(CE_INFO, "Disconnected from %s", rig_dev);
	else
		cmn_err(CE_INFO, "Failed to close rig: %s", xstrerror(ret));
}

void *rig_watcher(void *private)
{
	for (;;) {
		struct hrig_state state;
		int ret;

		if (!rig_connect()) {
			sleep(RECONNECT_DELAY);
			continue;
		}

		ret = hamlib_get_state(rig, &state);
		if (ret) {
			/* let the next iteration reconnect */
			rig_disconnect();
			continue;
		}

		notify(ANNOUNCE_RIG_FREQ, state.tx_freq, state.rx_freq,
		       state.tx_mode, state.rx_mode, state.power);

		sleep(1);
	}

	cmn_err(CE_INFO, "Thread exiting...");

	notify(ANNOUNCE_RIG_DISCONNECT, 0, 0, 0, 0, 0);

	return NULL;
}

int rig_start(uint32_t model, const char *dev,
	      const char *lhost, uint16_t lport)
{
	int ret;

	if (!dev) {
		cmn_err(CE_INFO, "Rig device not specified, rig control "
			"not available");
		return 0; /* nothing to do */
	}

	rig_model = model;
	rig_dev   = dev;

	rig = hamlib_alloc(model, dev);
	if (IS_ERR(rig)) {
		cmn_err(CE_ERROR, "Failed to open rig: %s",
			xstrerror(PTR_ERR(rig)));
		return PTR_ERR(rig);
	}

	ret = xthr_create_named(&rig_thread, "rig",
				rig_watcher, NULL);
	if (ret)
		cmn_err(CE_ERROR, "Failed to start rig thread: %s",
			xstrerror(ret));

	if (lhost || lport) {
		struct rigctl_args *args;

		args = malloc(sizeof(struct rigctl_args));
		if (args) {
			args->rig = rig;
			args->lhost = lhost;
			args->lport = lport;

			ret = xthr_create_named(&rigctl_thread, "rigctl",
						rigctl_main, args);
			if (ret) {
				cmn_err(CE_ERROR, "Failed to start rigctl main thread: "
					"%s", xstrerror(ret));
				free(args);
			}
		}
	}

	return ret;
}

void rig_stop(void)
{
	int ret;

	if (!rig_dev)
		return; /* never started */

	ret = xthr_join(rig_thread, NULL);
	if (ret)
		cmn_err(CE_ERROR, "Failed to join rig thread: %s",
			xstrerror(ret));

	/* TODO: shut down socksvc */

	ret = xthr_join(rigctl_thread, NULL);
	if (ret)
		cmn_err(CE_ERROR, "Failed to join rigctl main thread: %s",
			xstrerror(ret));
}
