/*
 * Copyright (c) 2022-2023 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

/*
 * This file implements a Flex 6300-style command set (and even responds to
 * the ID command pretending to be a 6300).  However, we define the behavior
 * of this virtual rig as:
 *
 *  - VFO A always receives
 *  - VFO A transmits when not split
 *  - VFO B transmits when split
 */

#include <jeffpc/types.h>
#include <jeffpc/error.h>
#include <jeffpc/io.h>
#include <jeffpc/socksvc.h>
#include <jeffpc/thread.h>
#include <jeffpc/buffer.h>

#include <hlog-rpc/hamlib.h>

#include "rig.h"
#include "rigctl.h"

struct command {
	const char *name;
	int (*f)(struct hrig *rig, int fd, const char *data, bool with_params);
	size_t nparam;
};

static int cmd_ai(struct hrig *rig, int fd, const char *data, bool with_params);
static int cmd_fa(struct hrig *rig, int fd, const char *data, bool with_params);
static int cmd_fb(struct hrig *rig, int fd, const char *data, bool with_params);
static int cmd_id(struct hrig *rig, int fd, const char *data, bool with_params);
static int cmd_if(struct hrig *rig, int fd, const char *data, bool with_params);
static int cmd_md(struct hrig *rig, int fd, const char *data, bool with_params);
static int cmd_rx(struct hrig *rig, int fd, const char *data, bool with_params);
static int cmd_tx(struct hrig *rig, int fd, const char *data, bool with_params);
static int cmd_zzfi(struct hrig *rig, int fd, const char *data, bool with_params);

static const struct command commands[] = {
	{ "AI", cmd_ai, 1, },
	{ "FA", cmd_fa, 11, },
	{ "FB", cmd_fb, 11, },
	{ "ID", cmd_id, 0, },
	{ "IF", cmd_if, 0, },
	{ "MD", cmd_md, 1, },
	{ "RX", cmd_rx, 0, },
	{ "TX", cmd_tx, 1, },
	{ "ZZFI", cmd_zzfi, 0, },
};

static enum hrig_mode flexmode2mode[] = {
	[1] = HRIG_MODE_LSB,
	[2] = HRIG_MODE_USB,
	[3] = HRIG_MODE_USB_CW,
	[4] = HRIG_MODE_FM,
	[5] = HRIG_MODE_AM,
	[6] = HRIG_MODE_LSB_DATA,
	[9] = HRIG_MODE_USB_DATA,
};

static int mode2flexmode[] = {
	[HRIG_MODE_UNKNOWN] = 0, /* extension */
	[HRIG_MODE_LSB] = 1,
	[HRIG_MODE_USB] = 2,
	[HRIG_MODE_USB_CW] = 3,
	[HRIG_MODE_FM] = 4,
	[HRIG_MODE_AM] = 5,
	[HRIG_MODE_LSB_DATA] = 6,
	[HRIG_MODE_USB_DATA] = 9,
};

struct bio {
	int fd;
	struct buffer buf;
};

static inline uint64_t get_vfo_freq(const struct hrig_state *state, char vfo)
{
	return (vfo == 'A') ? state->rx_freq : state->tx_freq;
}

static void log_msg(char dir, const void *data, size_t len)
{
	cmn_err(CE_DEBUG, "%c: %*.*s", dir, (int) len, (int) len, (char *) data);
}

static int log_and_write(int fd, const char *data, size_t len)
{
	log_msg('S', data, len);

	return xwrite(fd, data, len);
}

static int cmd_ai(struct hrig *rig, int fd, const char *data, bool with_params)
{
	const char out[4] = "AI0;";

	if (with_params) {
		switch (data[2]) {
			case '0':
				return 0;
			default:
				return -ENOTSUP;
		}
	} else {
		return xwrite(fd, out, sizeof(out));
	}
}

static int cmd_f(struct hrig *rig, int fd, const char *data, bool with_params,
		 char vfo)
{
	char out[14 + 1];
	int ret;

	if ((vfo != 'A') && (vfo != 'B'))
		return -ENOTSUP;

	if (with_params) {
		const bool rx = vfo == 'A';
		uint64_t freq;

		ret = str2u64_full(&data[2], &freq, 10, ';');
		if (ret)
			return ret;

		return hamlib_set_freq(rig, freq, rx);
	} else {
		struct hrig_state state;

		ret = hamlib_get_state(rig, &state);
		if (ret)
			return ret;

		snprintf(out, sizeof(out), "F%c%011"PRIu64";", vfo,
			 (vfo == 'A') ? state.rx_freq : state.tx_freq);

		return xwrite(fd, out, sizeof(out) - 1);
	}
}

static int cmd_fa(struct hrig *rig, int fd, const char *data, bool with_params)
{
	return cmd_f(rig, fd, data, with_params, 'A');
}

static int cmd_fb(struct hrig *rig, int fd, const char *data, bool with_params)
{
	return cmd_f(rig, fd, data, with_params, 'B');
}

static int cmd_id(struct hrig *rig, int fd, const char *data, bool with_params)
{
	const char out[6] = "ID907;"; /* pretend we're Flex-6300 */

	if (with_params)
		return -ENOTSUP; /* not allowed by the standard */

	return xwrite(fd, out, sizeof(out));
}

static int cmd_if(struct hrig *rig, int fd, const char *data, bool with_params)
{
	struct hrig_state state;
	char out[38 + 1];
	int ret;

	if (with_params)
		return -ENOTSUP; /* not allowed by the standard */

	ret = hamlib_get_state(rig, &state);
	if (ret)
		return ret;

	snprintf(out, sizeof(out), "IF%011"PRIu64"%04u%c%05u%u%u%u%02u%u%u%u%u%u%u%02u%u;",
		 get_vfo_freq(&state, 'A'), /* VFO A freq */
		 1 /* TODO: step size */,
		 '+', 0, /* TODO: rit/xit freq */
		 0, /* TODO: rit status */
		 0, /* TODO: xit status */
		 0, /* channel bank number (unused) */
		 0, /* channel bank number (unused) */
		 state.ptt ? 1 : 0,
		 mode2flexmode[state.rx_mode], /* VFO A mode */
		 0, /* same as FR */
		 0, /* split status (unused) */
		 state.split ? 1 : 0, /* same as FT */
		 0, /* CTCSS tone (unused) */
		 0, /* more tone control (unused) */
		 0); /* shift status (unused) */

	return xwrite(fd, out, sizeof(out) - 1);
}

static int cmd_md(struct hrig *rig, int fd, const char *data, bool with_params)
{
	struct hrig_state state;
	char out[4 + 1];
	int ret;

	if (with_params) {
		unsigned char flexmode = data[2] - '0';

		if ((flexmode >= ARRAY_LEN(flexmode2mode)) ||
		    !flexmode2mode[flexmode])
			return -EINVAL;

		return hamlib_set_mode(rig, flexmode2mode[flexmode], true);
	}

	ret = hamlib_get_state(rig, &state);
	if (ret)
		return ret;

	snprintf(out, sizeof(out), "MD%u;", mode2flexmode[state.tx_mode]);

	return xwrite(fd, out, sizeof(out) - 1);
}

/* write-only command without args */
static int cmd_rx(struct hrig *rig, int fd, const char *data, bool with_params)
{
	if (with_params)
		return -ENOTSUP; /* not allowed by the standard */

	return hamlib_set_ptt(rig, false); /* force PTT off */
}

/* write-only command with optional arg */
static int cmd_tx(struct hrig *rig, int fd, const char *data, bool with_params)
{
	if (with_params)
		return -EINVAL; /* TODO */

	return hamlib_set_ptt(rig, true); /* force PTT on */
}

static int cmd_zzfi(struct hrig *rig, int fd, const char *data, bool with_params)
{
	const char out[7] = "ZZFI00;";

	if (with_params)
		return 0; /* FIXME */

	return xwrite(fd, out, sizeof(out));
}

static int __buffer_read_append(struct bio *bio, size_t len)
{
	char tmp[len];
	int ret;

	ret = xread(bio->fd, tmp, len);
	if (ret)
		return ret;

	VERIFY0(buffer_append(&bio->buf, tmp, len));

	return 0;
}

/* read in enough so that the internal buffer contains a full command */
static int run_command(struct hrig *rig, struct bio *bio)
{
	struct buffer *buf = &bio->buf;
	const struct command *cmd;
	bool with_params;
	size_t cmdlen;
	size_t i;
	int ret;

	/* need full command name to decide length */
	if (buffer_size(buf) < 2) {
		ret = __buffer_read_append(bio, 2 - buffer_size(buf));
		if (ret)
			return ret;
	}

	/* ZZ commands use 4 char names */
	if (!memcmp("ZZ", buffer_data(buf), 2) &&
	    (buffer_size(buf) < 4)) {
		ret = __buffer_read_append(bio, 4 - buffer_size(buf));
		if (ret)
			return ret;
	}

	/* identify the command */
	for (i = 0; i < ARRAY_LEN(commands); i++) {
		cmd = &commands[i];

		if (!memcmp(cmd->name, buffer_data(buf),
			    strlen(cmd->name)))
			break;
	}

	if (i == ARRAY_LEN(commands)) {
		/*
		 * unknown command - read until ';' in hopes of
		 * resynchronizing with the sender (FIXME)
		 */
		log_msg('C', buffer_data(buf), buffer_size(buf));
		return log_and_write(bio->fd, "?;", 2);
	}

	cmdlen = strlen(cmd->name) + 1;
	ASSERT3U(cmdlen, <=, buf->allocsize /* FIXME */);

	/* read in enough to get a full get command */
	if (buffer_size(buf) < cmdlen) {
		ret = __buffer_read_append(bio, cmdlen - buffer_size(buf));
		if (ret)
			return ret;
	}

	with_params = ((char *) buffer_data(buf))[strlen(cmd->name)] != ';';

	/* if there are params, buffer them */
	if (with_params) {
		cmdlen += cmd->nparam;
		ASSERT3U(cmdlen, <=, buf->allocsize /* FIXME */);

		/* read in enough for the command itself */
		ret = __buffer_read_append(bio, cmdlen - buffer_size(buf));
		if (ret)
			return ret;
	}

	/* dispatch command */
	ret = cmd->f(rig, bio->fd, buffer_data(buf),
		     (cmdlen - strlen(cmd->name)) > 1);
	if (ret) {
		cmn_err(CE_DEBUG, "Failing command: '%*.*s'",
			(int) cmdlen, (int) cmdlen, (char *) buffer_data(buf));
	}

	/* remove the command from the buffer */
	buffer_delete(buf, 0, cmdlen);

	return ret;
}

static void rigctl_process(int fd, struct socksvc_stats *stats, void *private)
{
	struct hrig *rig = private;
	struct bio bio;
	size_t maxsize;
	size_t i;

	xthr_set_name("rigctl");
	cmn_err(CE_DEBUG, "got connection %d", fd);

	/* find the longest supported command */
	maxsize = 0;
	for (i = 0; i < ARRAY_LEN(commands); i++) {
		const struct command *cmd = &commands[i];

		maxsize = MAX(maxsize, strlen(cmd->name) + cmd->nparam + 1);
	}

	/* set up a buffer */
	maxsize *= 2;
	bio.fd = fd;
	buffer_init_static(&bio.buf, alloca(maxsize), 0, maxsize, true);

	/* process commands */
	for (;;) {
		int ret;

		ret = run_command(rig, &bio);
		if (ret < 0) {
			cmn_err(CE_INFO, "Failed to read from socket: %s",
				xstrerror(ret));
			break;
		}
	}

	cmn_err(CE_INFO, "Closing connection %d", fd);
}

void *rigctl_main(void *private)
{
	struct rigctl_args *args = private;
	int ret;

	ret = socksvc(args->lhost, args->lport, -1, rigctl_process, args->rig);
	if (ret)
		cmn_err(CE_ERROR, "Failed to start rigctl server: %s",
			xstrerror(ret));

	free(args);

	return NULL;
}
