/*
 * Copyright (c) 2021-2022,2024 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <hlog/qso.h>
#include <hlog/index.h>
#include <hlog-rpc/announce.h>
#include <hlog-rpc/announce-pub.h>

#include "contest.h"

static void print_notification(enum announce_qso_msg_type type,
			       const char *source, const char *uuid,
			       struct str *call, struct str *mode,
			       struct str *grid)
{
	switch (type) {
		case ANNOUNCE_QSO_NEW:
		case ANNOUNCE_QSO_UPDATED: {
			const char *action = (type == ANNOUNCE_QSO_NEW) ?
				"added" : "updated";

			msg_print(MSG_QSO, CE_INFO,
				  "QSO %-7s %s %s %s %s (%s)",
				  action, uuid, str_cstr(call), str_cstr(mode),
				  str_cstr(grid), source);
			break;
		}
	}

	str_putref(call);
	str_putref(mode);
	str_putref(grid);
}

int save_qso_and_update_index(struct qso *qso, const char *source)
{
	int ret;

	/*
	 * First, save the QSO entry to disk
	 */
	ret = qso_save(qso, AT_FDCWD);
	if (ret) {
		msg_print(MSG_QSO, CE_ERROR, "Failed to save QSO: %s",
			  xstrerror(ret));
		return ret;
	}

	/*
	 * Let everyone know about the update
	 */
	print_notification(ANNOUNCE_QSO_NEW, source, qso->uuid_str,
			   qso_get_station_call(&qso->rx),
			   qso_get_mode(&qso->rx),
			   qso_get_grid(&qso->rx));

	announce_publish_qso(ANNOUNCE_QSO_NEW, &qso->uuid);

	/*
	 * Now, update the index
	 */
	ret = index_save(qso);
	if (ret)
		msg_print(MSG_QSO, CE_WARN, "Failed to add QSO to index, "
			  "UI may be out-of-sync: %s", xstrerror(ret));

	return ret;
}
