/*
 * Copyright (c) 2021 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/types.h>
#include <jeffpc/cstr.h>

#include <hlog-rpc/announce.h>

void announce_pack_loc_msg(struct announce_msg *msg,
			   enum announce_loc_msg_type type, uint8_t mode,
			   const char *grid)
{
	/* make sure mode and grid match and that mode is valid */
	switch (mode) {
		case 0:
			if (grid && strlen(grid))
				goto err_inval_grid;
			grid = "";
			break;
		case 2:
		case 3:
			if (!grid || !strlen(grid) || (strlen(grid) % 2 != 0))
				goto err_inval_grid;
			break;
		default:
			cmn_err(CE_WARN, "Ignoring location notification "
				"with invalid mode=%u", mode);
			return;
	}

	/* make sure the message type and mode match */
	switch (type) {
		case ANNOUNCE_LOC_ACQ:
			if ((mode != 2) && (mode != 3))
				goto err_inval_mode;
			break;
		case ANNOUNCE_LOC_CHANGE:
		case ANNOUNCE_LOC_CURRENT:
			/* any of the modes is ok */
			break;
		case ANNOUNCE_LOC_LOST:
		case ANNOUNCE_LOC_DISCONNECT:
			if (mode != 0)
				goto err_inval_mode;
			break;
	}

	msg->major = cpu8_to_be(ANNOUNCE_LOC);
	msg->minor = cpu8_to_be(type);
	msg->u.loc.mode = cpu8_to_be(mode);
	strcpy_safe((char *) msg->u.loc.grid, grid, sizeof(msg->u.loc.grid));

	return;

err_inval_grid:
	cmn_err(CE_WARN, "Ignoring location notification with invalid grid "
		"(msg %d, mode %d, grid %p, len %zu)", type, mode, grid,
		grid ? strlen(grid) : 0);
	return;

err_inval_mode:
	cmn_err(CE_WARN, "Ignoring location notification with invalid mode "
		"(msg %d, mode %d)", type, mode);
}
