/*
 * Copyright (c) 2021,2024 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/error.h>
#include <jeffpc/types.h>
#include <jeffpc/cstr.h>
#include <jeffpc/time.h>

#include <hlog/udpsvc.h>

#include <hlog-rpc/announce.h>
#include <hlog-rpc/announce-pub.h>

static struct udpsvc *svc;

static struct {
	void (*qso)(enum announce_qso_msg_type, struct xuuid *);
	void (*loc)(enum announce_loc_msg_type, int mode, const char *grid);
	void (*rig)(enum announce_rig_msg_type, uint64_t tx_freq,
		    uint64_t rx_freq, enum hrig_mode tx_mode,
		    enum hrig_mode rx_mode, uint32_t power);
} notify;

static void interpret_msg(struct announce_msg *msg)
{
	/* byte-order */
	msg->magic = be32_to_cpu(msg->magic);
	msg->major = be8_to_cpu(msg->major);
	msg->minor = be8_to_cpu(msg->minor);
	msg->ts = be64_to_cpu(msg->ts);

	switch ((enum announce_msg_type) msg->major) {
		case ANNOUNCE_QSO:
			break; /* nothing else to byte-order */
		case ANNOUNCE_LOC:
			msg->u.loc.mode = be8_to_cpu(msg->u.loc.mode);
			break;
		case ANNOUNCE_RIG:
			msg->u.rig.tx_freq = be64_to_cpu(msg->u.rig.tx_freq);
			msg->u.rig.rx_freq = be64_to_cpu(msg->u.rig.rx_freq);
			msg->u.rig.power = be32_to_cpu(msg->u.rig.power);
			msg->u.rig.tx_mode = be8_to_cpu(msg->u.rig.tx_mode);
			msg->u.rig.rx_mode = be8_to_cpu(msg->u.rig.rx_mode);
			break;
	}

	/* actually interpret */
	if (msg->magic != ANNOUNCE_CURRENT_VERSION)
		return; /* ignore non-announcements */

	switch ((enum announce_msg_type) msg->major) {
		case ANNOUNCE_QSO:
			if (notify.qso) {
				struct xuuid uuid;

				memcpy(uuid.raw, msg->u.qso.uuid, sizeof(uuid));

				notify.qso(msg->minor, &uuid);
			}
			break;
		case ANNOUNCE_LOC:
			if (notify.loc)
				notify.loc(msg->minor, msg->u.loc.mode,
					   (const char *) msg->u.loc.grid);
			break;
		case ANNOUNCE_RIG:
			if (notify.rig)
				notify.rig(msg->minor, msg->u.rig.tx_freq,
					   msg->u.rig.rx_freq,
					   msg->u.rig.tx_mode,
					   msg->u.rig.rx_mode,
					   msg->u.rig.power);
			break;
	}
}

static void process(void *buf, size_t nbytes, void *private)
{
	struct announce_msg msg;

	if (nbytes != sizeof(struct announce_msg))
		return; /* ignore messages of wrong size */

	memcpy(&msg, buf, sizeof(msg));

	interpret_msg(&msg);
}

int announce_start(void)
{
	svc = udpsvc(NULL, ANNOUNCE_PORT, true, "announce", process, NULL);
	if (IS_ERR(svc))
		return PTR_ERR(svc);

	return 0;
}

void announce_stop(void)
{
	udpsvc_stop(svc);
}

int announce_connect(void (*qso)(enum announce_qso_msg_type, struct xuuid *),
		     void (*loc)(enum announce_loc_msg_type, int mode,
				 const char *grid),
		     void (*rig)(enum announce_rig_msg_type, uint64_t tx_freq,
				 uint64_t rx_freq, enum hrig_mode tx_mode,
				 enum hrig_mode rx_mode, uint32_t power))
{
	notify.qso = qso;
	notify.loc = loc;
	notify.rig = rig;

	return 0;
}

void announce_publish(struct announce_msg *msg)
{
	int ret;

	/*
	 * Don't trust that magic is set and padding is zero filled
	 */
	msg->magic = cpu32_to_be(ANNOUNCE_CURRENT_VERSION);
	msg->_pad0 = 0;
	msg->ts = cpu64_to_be(gettime());
	memset(msg->_pad1, 0, sizeof(msg->_pad1));

	switch ((enum announce_msg_type) be8_to_cpu(msg->major)) {
		case ANNOUNCE_QSO:
			memset(msg->u.qso._pad2, 0, sizeof(msg->u.qso._pad2));
			break;
		case ANNOUNCE_LOC:
			memset(msg->u.loc._pad2, 0, sizeof(msg->u.loc._pad2));
			break;
		case ANNOUNCE_RIG:
			memset(msg->u.rig._pad2, 0, sizeof(msg->u.rig._pad2));
			break;
	}

	/*
	 * Publish
	 */
	ret = udpsvc_bcast(svc, msg, sizeof(*msg));
	if (ret)
		cmn_err(CE_WARN, "announce: failed to send notification: %s",
			strerror(errno));
}
