/*
 * Copyright (c) 2021,2023-2024 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#ifndef __RPC_ANNOUNCE_H
#define __RPC_ANNOUNCE_H

#include <jeffpc/error.h>
#include <jeffpc/uuid.h>

#include <hlog-rpc/rig-mode.h>

/*
 * The announcement message queue is a PUB/SUB queue the daemon binds to and
 * the clients connect to.  The daemon then publishes a message every time
 * the QSO index changes, the station location changes, etc.
 *
 * The notification message (see struct announce_msg) is very simple.  It
 * consists of a pair of 8-bit message types (major and minor) followed by
 * the 24-byte payload containing more information about the notification.
 */

#define ANNOUNCE_PORT		7373

enum announce_msg_type {
	ANNOUNCE_QSO = 0x1,
	ANNOUNCE_LOC,
	ANNOUNCE_RIG,
};

enum announce_qso_msg_type {
	ANNOUNCE_QSO_NEW = 0x0,
	ANNOUNCE_QSO_UPDATED,
};

enum announce_loc_msg_type {
	ANNOUNCE_LOC_CURRENT = 0x0,
	ANNOUNCE_LOC_ACQ,
	ANNOUNCE_LOC_CHANGE,
	ANNOUNCE_LOC_LOST,
	ANNOUNCE_LOC_DISCONNECT, /* device disconnected */
};

enum announce_rig_msg_type {
	ANNOUNCE_RIG_FREQ = 0x0,
	ANNOUNCE_RIG_DISCONNECT, /* device disconnected */
};

#define ANNOUNCE_CURRENT_VERSION	0x686c6700 /* "hlg" & 0 */
struct announce_msg {
	uint32_t magic; /* also serves as version */
	uint8_t major;
	uint8_t minor;
	uint16_t _pad0;
	uint64_t ts;
	uint8_t _pad1[48];

	union {
		struct {
			uint8_t uuid[16];
			uint8_t _pad2[112];
		} qso; /* ANNOUNCE_QSO */
		struct {
			uint8_t mode; /* 0 = no fix, 2 = 2D, 3 = 3D */
			uint8_t grid[9]; /* nul terminated */
			uint8_t _pad2[118];
		} loc; /* ANNOUNCE_LOC */
		struct {
			uint64_t tx_freq;
			uint64_t rx_freq;
			uint32_t power;
			uint8_t tx_mode;
			uint8_t rx_mode;
			uint8_t _pad2[106];
		} rig; /* ANNOUNCE_RIG */
	} u;
} __attribute__((packed));
STATIC_ASSERT(sizeof(struct announce_msg) == (64 + 128));

/* fill in the msg pointer with a properly formed announcement */
extern void announce_pack_loc_msg(struct announce_msg *msg,
				  enum announce_loc_msg_type type, uint8_t mode,
				  const char *grid);
extern void announce_pack_qso_msg(struct announce_msg *msg,
				  enum announce_qso_msg_type type,
				  const struct xuuid *uuid);
extern void announce_pack_rig_msg(struct announce_msg *msg,
				  enum announce_rig_msg_type type,
				  uint64_t tx_freq, uint64_t rx_freq,
				  enum hrig_mode tx_mode, enum hrig_mode rx_mode,
				  uint32_t power);

#endif
