/*
 * Copyright (c) 2020-2021 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/version.h>
#include <jeffpc/error.h>
#include <jeffpc/io.h>
#include <jeffpc/val.h>
#include <jeffpc/rbtree.h>

#include <hlog/version.h>
#include <hlog/qso.h>
#include <hlog/index.h>
#include <hlog/util.h>
#include <hlog-lua/xlua.h>

static const char *data_dir;
static int data_dir_fd;

static void process(const char *scriptname)
{
	char error[128];
	lua_State *L;
	int ret;

	L = xlua_init(scriptname, 0, error, sizeof(error));
	if (IS_ERR(L)) {
		printf("Error: failed to initialize lua: %s\n", error);
		return;
	}

	ret = xlua_call(L, "main", 0, 0, error, sizeof(error));
	if (ret)
		printf("Error: failed to execute lua: %s\n", error);

	lua_close(L);
}

static void usage(const char *prog)
{
	fprintf(stderr, "HLog version %s (libjeffpc %s)\n\n", version_string,
		jeffpc_version);

	fprintf(stderr, "Usage: %s <lua script file name> <path to data dir>\n",
		prog);
	exit(1);
}

static int load_index(void)
{
	int ret;

	printf("Loading existing QSOs...\n");
	ret = index_init(data_dir_fd);
	printf("%s.\n", ret ? "error" : "done");

	if (ret)
		fprintf(stderr, "Error: failed to load existing entries: %s\n",
			xstrerror(ret));

	return ret;
}

int main(int argc, char **argv)
{
	const char *script;
	int ret;

	if (argc != 3)
		usage(argv[0]);

	script = argv[1];

	data_dir = argv[2];
	data_dir_fd = xopen(data_dir, O_RDONLY, 0);
	if (data_dir_fd < 0) {
		fprintf(stderr, "Error: failed to open data directory: %s\n",
			xstrerror(data_dir_fd));
		goto err;
	}

	ret = load_index();
	if (ret)
		goto err_close;

	process(script);

	xclose(data_dir_fd);

	return 0;

err_close:
	xclose(data_dir_fd);

err:
	return 1;
}
