--
-- Copyright (c) 2022 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
-- SOFTWARE.
--

local ALLOWED_BANDS = {
	[2000] = true,
	[6000] = true,
}

local dup_table = {}
local mults = {}

for band, _ in pairs(ALLOWED_BANDS) do
	dup_table[band] = {}
end

local function mkgrid(g)
	if g == nil then
		return ""
	end

	return g:sub(1,4)
end

local function hash(mgrid, tgrid)
	return string.format("%s-%s", mkgrid(mgrid), mkgrid(tgrid))
end

local function check(band, who, mgrid, tgrid)
	local where = hash(mgrid, tgrid)
	local info = dup_table[band][who]

	if info == nil then
		return 1 -- callsign not seen before on this band ==> not dup
	end

	if info[where] then
		return 3 -- exact match ==> dup
	end

	for cur, _ in pairs(info) do
		if cur:find(where, 1, true) == 1 then
			return 2 -- where is a prefix of cur ==> maybe dup
		end
	end

	return 1 -- no prefix found ==> not dup
end

local function qso_done(qso)
	if not ALLOWED_BANDS[qso.tx.band] then
		return -- ignore out-of-band contacts
	end

	qso.additional['contest-id'] = 'CQ-WW-VHF'
	qso.additional['contest-year'] = contest.year

	local where = hash(qso.tx.grid, qso.rx.grid)
	local mult = string.format("%u-%s", qso.tx.band, where)

	if dup_table[qso.tx.band][qso.rx.station_call] == nil then
		dup_table[qso.tx.band][qso.rx.station_call] = {}
	end
	dup_table[qso.tx.band][qso.rx.station_call][where] = true

	mults[mult] = true
end

--
-- Annunciator color
--
-- There are three colors we can display: "not a dup", "maybe a dup", and
-- "definitely a dup".  When we have full information (i.e., callsign and
-- grid), it is easy to check and return either "not a dup" or "definitely a
-- dup".  Things get more complicated when we don't have all the info.  In
-- general, we try to do a prefix match with the entered grid square.  This
-- allows us to indicate whether or not there is some chance of a dup.
local function is_dup(qso, band, color_none, color_warn, color_dup)
	local colors = { color_none, color_warn, color_dup }
	local who = qso.rx.station_call
	local where = qso.rx.grid

	--
	-- complete information
	--
	if where ~= nil and where:len() >= 4 then
		local colors = { color_none, nil, color_dup }

		return colors[check(band, who, qso.tx.grid, where)]
	end

	--
	-- incomplete information
	--
	if where ~= nil then
		return colors[check(band, who, qso.tx.grid, where)]
	end

	return colors[check(band, who, qso.tx.grid, "")]
end

local function field_changed(qso, fields, field)
	local value = fields[field]

	if field == "rx.station_call" then
		qso.rx.station_call = value
	elseif field == "rx.grid" then
		qso.rx.grid = value
	elseif field == "comment" then
		qso.comment = value
	else
		error(string.format("Unknown field '%s' encountered", field))
	end

	if not ALLOWED_BANDS[qso.tx.band] then
		return -- ignore out-of-band contacts
	end

	return nil, {
		["2m"] = is_dup(qso, 2000, "green", "magenta", "off"),
		["6m"] = is_dup(qso, 6000, "green", "magenta", "off"),
		["DUP"] = is_dup(qso, qso.tx.band, "black", "black", "red"),
	}
end

return {
	labels = {
		{ 0,  0, "St. Call" },
		{ 0, 11, "Grid" },
		{ 0, 20, "Comment" },
	},

	fields = {
		{ 1,  0, 10, "rx.station_call", },
		{ 1, 11,  8, "rx.grid", },
		{ 1, 20, 20, "comment", },
	},

	annunciators = {
		{ 0, 31, "DUP", "black" },
		{ 0, 35, "6m", "black" },
		{ 0, 38, "2m", "black" },
	},

	events = {
		qso_init = function(qso)
			contest.lcd(mkgrid(qso.tx.grid))
		end,
		qso_done = qso_done,
		field_changed = field_changed,
	},
}
