--
-- Copyright (c) 2020-2021,2024 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
-- SOFTWARE.
--

local totals = {}
local bands = {}
local band_names = {}

local function init_mode_band(mode, band)
	if totals[mode] == nil then
		totals[mode] = 0
	end

	if bands[mode] == nil then
		bands[mode] = {}
	end

	if bands[mode][band] == nil then
		bands[mode][band] = 0
	end

	band_names[band] = band
end

function summarize_qso(qso)
	local mode = qso.tx.mode
	local band = qso.tx.band

	if mode == nil then
		mode = "<Unknown>"
	end

	if band == nil then
		band = 0
	end

	init_mode_band(mode, band)

	-- increment
	totals[mode] = totals[mode] + 1
	bands[mode][band] = bands[mode][band] + 1
end

local function band_cmp(a, b)
	return a > b
end

local function mode_cmp(a, b)
	-- we want to sort based per-mode count (desc), not the name
	if totals[a] ~= totals[b] then
		return totals[a] > totals[b]
	end

	-- if per-mode count is the same, sort by mode name (asc)
	return a < b
end

function print_summary()
	local out

	-- order the band names
	local band_name_array = {}
	for band, dummy in pairs(band_names) do
		table.insert(band_name_array, band)
	end

	table.sort(band_name_array, band_cmp)

	-- order the mode names
	local mode_name_array = {}
	local total = 0
	for mode, count in pairs(totals) do
		table.insert(mode_name_array, mode)
		total = total + count
	end

	table.sort(mode_name_array, mode_cmp)

	-- print header
	out = string.format("%-10s %4s ", "", "#")
	for idx, band in pairs(band_name_array) do
		if band == 0 then
			out = out.."U/O "
		elseif band > 10000 then
			out = out..string.format("%3.0f ", band / 1000.0)
		elseif band % 1000 ~= 0 then
			out = out..string.format("%3.1f ", band / 1000.0)
		else
			out = out..string.format("%3.0f ", band / 1000.0)
		end
	end
	print(out)

	-- print each mode stats
	for idx, mode in pairs(mode_name_array) do
		out = string.format("%-10s %4u ", string.sub(mode, 1, 10),
				    totals[mode])

		for idx, band in pairs(band_name_array) do
			if bands[mode][band] then
				out = out.." X  "
			else
				out = out.."    "
			end
		end

		print(out)
	end

	-- print totals
	print("---------------")
	print(string.format("%-10s %4u", "Total", total))
end
