/*
 * Copyright (c) 2022 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include "xlua_impl.h"
#include <hlog/sig.h>

#define _S(n, f, v) \
	do { \
		f(L, v); \
		lua_setfield(L, -2, n); \
	} while (0)
#define S(n, x) \
	do { \
		if (x) \
			_S(n, lua_pushstring, str_cstr(x)); \
	} while (0)
#define N(n, x) _S(n, lua_pushnumber, x)
#define B(n, x) _S(n, lua_pushboolean, x)
#define I(n, x, allowzero) \
	do { \
		if (allowzero || (x)) \
			_S(n, lua_pushinteger, x); \
	} while (0)

/* return a table with POTA info or nil [-1,+1,-] */
static int xlua_pota_lookup(lua_State *L)
{
	struct sig_pota_ref info;
	const char *ref;

	xlua_checkstack(L, 1, 1, 1);

	ref = luaL_checkstring(L, -1);

	/*
	 * First, get info
	 */
	if (sig_lookup_pota(ref, &info)) {
		lua_pushnil(L);
		return 1;
	}

	/*
	 * Second, stuff it into a table
	 */
	lua_newtable(L);

	S("ref", info.ref);
	S("name", info.name);
	S("locs", info.locs);
	B("active", info.active);
	I("dxcc", info.dxcc, true);

	if (info.have_latlon) {
		N("lat", info.lat);
		N("lon", info.lon);
	}

	str_putref(info.ref);
	str_putref(info.name);
	str_putref(info.locs);

	return 1;
}

/* return a table with SOTA info or nil [-1,+1,-] */
static int xlua_sota_lookup(lua_State *L)
{
	struct sig_sota_ref info;
	const char *ref;

	xlua_checkstack(L, 1, 1, 1);

	ref = luaL_checkstring(L, -1);

	/*
	 * First, get info
	 */
	if (sig_lookup_sota(ref, &info)) {
		lua_pushnil(L);
		return 1;
	}

	/*
	 * Second, stuff it into a table
	 */
	lua_newtable(L);

	S("ref", info.ref);
	S("name", info.name);
	I("points", info.points, true);

	str_putref(info.ref);
	str_putref(info.name);

	return 1;
}

static const luaL_Reg sig_funcs[] = {
	{ "lookup_pota", xlua_pota_lookup },
	{ "lookup_sota", xlua_sota_lookup },
	{ NULL, NULL },
};

/* register the sig library [-0,+0,e] */
void xlua_pushlib_sig(lua_State *L)
{
	luaL_newlib(L, sig_funcs);
}
