/*
 * Copyright (c) 2020-2021 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#ifndef __WRAPPER_H
#define __WRAPPER_H

/*
 * The following macro generates everything necessary to wrap a C structure
 * and easily push it onto Lua stacks as well as check that a value on the
 * stack is of this type.
 *
 * To generate the necessary API, simply use the macro in your .c file.  The
 * arguments consist of:
 *
 *   <typename> = the name for the structure to use in generated symbols
 *   <type> = the exposed structure's C type
 *   <putref> = the function to remove a reference from the C structure,
 *       called from the automatically generated __gc metamethod
 *   <extras> = a C array of luaL_Reg structs passed to luaL_setfuncs to
 *       add Lua methods to the type
 *
 * For example, to expose struct timespec with a tostring method, one would
 * use the following in a .c file:
 *
 *   // struct timespec isn't refcounted
 *   static inline void timespec_putref(struct timespec *ts)
 *   {
 *   }
 *
 *   static int timespec_tostring(lua_State *L);
 *
 *   static const luaL_Reg timespec_fxns[] = {
 *           { "__tostring", timespec_tostring },
 *           { NULL, NULL },
 *   };
 *
 *   WRAP_C_DATA(timespec, struct timespec, timespec_putref, timespec_fxns)
 *
 *   static int timespec_tostring(lua_State *L)
 *   {
 *           struct timespec *ts;
 *
 *           ts = xlua_timespec_getptr(L, -1, NULL);
 *
 *           lua_pushfstring(L, "%d.%09ld", ts->tv_sec, ts->tv_nsec);
 *
 *           return 1;
 *   }
 *
 *
 * The two functions making up the exposed structure API are:
 *
 *   void xlua_push<typename>(lua_State *, <type> *, bool);
 *   static inline <type> *xlua_<typename>_getptr(lua_State *, int, bool *);
 *
 * The first pushes the supplied pointer onto the Lua stack, while the
 * second returns the previously pushed C pointer (or generates a Lua error
 * if there is a type mismatch).
 *
 * The push function tracks the passed in bool argument alongside the C
 * pointer itself.  The getptr function fills in the bool pointer (if not
 * NULL) with the pushed bool value.  While this bool isn't used directly by
 * the generated code and therefore it can be used for anything, the
 * intention is to allow the API consumers to mark certain pointers as
 * "read-only" to prevent Lua scripts from attempting to mutate them.  This
 * allows one to push const pointers while "notifying" the methods that
 * mutation of any of the members is undefined behavior.
 *
 * Also note that the getptr function is static inline.  That is, it is
 * intended to only be used in the file where it is generated.  It is a
 * somewhat low-level function.  If some form of a get function is desired,
 * it is up to the API consumer to write a wrapper around getptr and expose
 * it to the rest of the codebase.
 *
 * Revisiting the struct timespec example above, the generated API would
 * be:
 *
 *   void xlua_pushtimespec(lua_State *, struct timespec *, bool);
 *   static inline struct timespec *xlua_timespec_getptr(lua_State *, int,
 *                                                       bool *);
 */
#define WRAP_C_DATA(typename, type, putref, extras)			\
struct lua_wrapper_##typename {						\
	type *ptr;							\
	bool ro;							\
};									\
									\
static inline type *xlua_##typename##_getptr(lua_State *L, int idx,	\
					 bool *ro_r)			\
{									\
	struct lua_wrapper_##typename *wrapper;				\
									\
	wrapper = luaL_checkudata(L, idx, #type);			\
									\
	if (ro_r)							\
		*ro_r = wrapper->ro;					\
									\
	return wrapper->ptr;						\
}									\
									\
static int xlua_wrapper_##typename##_gc(lua_State *L)			\
{									\
	putref(xlua_##typename##_getptr(L, -1, NULL));			\
									\
	return 0;							\
}									\
									\
static const luaL_Reg provided_##typename##_fxns[] = {			\
	{ "__gc", xlua_wrapper_##typename##_gc },			\
	{ NULL, NULL },							\
};									\
									\
/* allocate a new wrapper struct [-0,+1,e] */				\
static int xlua_push##typename##_alloc(lua_State *L)			\
{									\
	const int orig_top = lua_gettop(L);				\
	struct lua_wrapper_##typename *wrapper;				\
									\
	wrapper = lua_newuserdata(L, sizeof(struct lua_wrapper_##typename)); \
	wrapper->ptr = NULL; /* putref(NULL) is safe in case we throw */ \
									\
	/* get the current metatable */					\
	if (luaL_getmetatable(L, #type) != LUA_TTABLE) {		\
		/* initialize a new metatable */			\
		const luaL_Reg *extra_fxns = (extras);			\
		bool have_index;					\
									\
		lua_pop(L, 1);						\
		luaL_newmetatable(L, #type);				\
		luaL_setfuncs(L, provided_##typename##_fxns, 0);	\
									\
		have_index = false;					\
		if (extra_fxns) {					\
			size_t i;					\
									\
			/* set the provided functions */		\
			luaL_setfuncs(L, extra_fxns, 0);		\
									\
			/* find __index, if provided */			\
			for (i = 0; extra_fxns[i].name; i++) {		\
				if (strcmp(extra_fxns[i].name, "__index")) \
					continue;			\
									\
				have_index = true;			\
				break;					\
			}						\
		}							\
									\
		if (!have_index) {					\
			/* set metatable["__index"] == the metatable */	\
			lua_pushliteral(L, "__index");			\
			lua_pushvalue(L, -2);				\
			lua_settable(L, -3);				\
		}							\
	}								\
									\
	/* set the metatable */						\
	lua_setmetatable(L, -2);					\
									\
	ASSERT_LUA_STACK(L, orig_top + 1); /* we pushed */		\
									\
	return 1;							\
}									\
									\
/* push [-0,+1,-] */							\
int xlua_push##typename(lua_State *L, type *ptr, bool ro, char *error,	\
			size_t errlen)					\
{									\
	int ret;							\
									\
	if (!lua_checkstack(L, 3)) {					\
		strcpy_safe(error, "can't grow lua stack - out of memory", \
			    errlen);					\
		return -ENOMEM;						\
	}								\
									\
	if (!ptr) {							\
		lua_pushnil(L);						\
		return 0;						\
	}								\
									\
	lua_pushcfunction(L, xlua_push##typename##_alloc);		\
									\
	ret = xlua_call(L, NULL, 0, 1, error, errlen);			\
	if (!ret) {							\
		struct lua_wrapper_##typename *wrapper;			\
									\
		wrapper = lua_touserdata(L, -1);			\
		wrapper->ptr = ptr;					\
		wrapper->ro = ro;					\
	} else {							\
		putref(ptr);						\
	}								\
									\
	return ret;							\
}									\
									\
/* push [-0,+1,e] */							\
void xlua_push##typename##_throws(lua_State *L, type *ptr, bool ro)	\
{									\
	char error[128];						\
	int ret;							\
									\
	ret = xlua_push##typename(L, ptr, ro, error, sizeof(error));	\
	if (ret)							\
		luaL_error(L, "%s", error);				\
}

#endif
